from mars_patcher.auto_generated_types import MarsschemaCreditstextItem
from mars_patcher.constants.credits_lines import (
    FUSION_STAFF_LINES,
    LINE_TYPE_HEIGHTS,
    LINE_TYPE_VALS,
    MARS_CREDITS,
    RDV_CREDITS,
    TEXT_LINE_TYPES,
    LineType,
)
from mars_patcher.rom import Rom

CREDITS_ADDR = 0x74B0B0
CREDITS_LEN = 0x2B98
FULL_LINE_LEN = 36
LINE_WIDTH = 30


class CreditsLine:
    LINE_TYPE_ENUMS = {
        "Blank": LineType.BLANK,
        "Blue": LineType.BLUE,
        "Red": LineType.RED,
        "White1": LineType.WHITE1,
        "White2": LineType.WHITE2,
    }

    def __init__(
        self,
        line_type: LineType,
        blank_lines: int = 0,
        text: str | None = None,
        centered: bool = True,
    ):
        self.line_type = line_type
        self.blank_lines = blank_lines
        self.text = text
        self.centered = centered

    @classmethod
    def from_json(cls, data: MarsschemaCreditstextItem) -> "CreditsLine":
        line_type = cls.LINE_TYPE_ENUMS[data["LineType"]]
        blank_lines = data.get("BlankLines", 0)
        text = data.get("Text")
        centered = data.get("Centered", True)
        return CreditsLine(line_type, blank_lines, text, centered)


def write_credits(rom: Rom, data: list[MarsschemaCreditstextItem]) -> None:
    writer = CreditsWriter(rom)
    # Write MARS credits
    lines = [CreditsLine(*line) for line in MARS_CREDITS]
    writer.write_lines(lines)
    # Write RDV credits
    lines = [CreditsLine(*line) for line in RDV_CREDITS]
    writer.write_lines(lines)
    # Write custom credits
    lines = [CreditsLine.from_json(d) for d in data]
    writer.write_lines(lines)
    # Write fusion staff credits
    lines = [CreditsLine(*line) for line in FUSION_STAFF_LINES]
    writer.write_lines(lines)


class CreditsWriter:
    def __init__(self, rom: Rom):
        self.rom = rom
        self.addr = CREDITS_ADDR
        self.num_lines = 0

    def write_lines(self, lines: list[CreditsLine]) -> None:
        for line in lines:
            lt_val = LINE_TYPE_VALS[line.line_type]
            line_bytes = bytearray([lt_val, line.blank_lines])
            if line.line_type in TEXT_LINE_TYPES and line.text:
                text = line.text
                if line.centered:
                    spacing = " " * ((LINE_WIDTH - len(text)) // 2)
                    text = spacing + text
                line_bytes.extend(text.encode("ascii"))
            self.write_line(line_bytes)
            line_height = LINE_TYPE_HEIGHTS.get(line.line_type, 0)
            self.num_lines += line_height + line.blank_lines

    def write_line(self, line: bytearray) -> None:
        if len(line) > FULL_LINE_LEN:
            raise ValueError(f"Line too long: {line}")
        self.rom.write_bytes(self.addr, line)
        for i in range(len(line), FULL_LINE_LEN):
            self.rom.write_8(self.addr + i, 0)
        self.addr += FULL_LINE_LEN



from mars_patcher.rom import Game, Region, Rom

# TODO: Consider moving these to JSON


def area_room_entry_ptrs(rom: Rom) -> int:
    """Returns the address of the area room entry pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79B8BC
        elif rom.region == Region.E:
            return 0x79C0F0
        elif rom.region == Region.J:
            return 0x7EDF6C
        elif rom.region == Region.C:
            return 0x77D5C0
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x75FAC4
        elif rom.region == Region.E:
            return 0x773964
        elif rom.region == Region.J:
            return 0x75FBD4
        elif rom.region == Region.C:
            return 0x79ECBC

    raise ValueError("Rom has unknown game loaded.")


def tileset_entries(rom: Rom) -> int:
    """Returns the address of the tileset entries."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x3BF888
        elif rom.region == Region.E:
            return 0x3BFEE4
        elif rom.region == Region.J:
            return 0x3C1E50
        elif rom.region == Region.C:
            return 0x3C1E94
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x33DFDC
        elif rom.region == Region.E:
            return 0x33EC68
        elif rom.region == Region.J:
            return 0x33E038
        elif rom.region == Region.C:
            return 0x3577C8

    raise ValueError("Rom has unknown game loaded.")


def tileset_count(rom: Rom) -> int:
    """Returns the number of tilesets in the game."""
    if rom.game == Game.MF:
        return 0x62
    elif rom.game == Game.ZM:
        return 0x4F
    raise ValueError(rom.game)


def area_doors_ptrs(rom: Rom) -> int:
    """Returns the address of the area doors pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79B894
        elif rom.region == Region.E:
            return 0x79C0C8
        elif rom.region == Region.J:
            return 0x7EDF44
        elif rom.region == Region.C:
            return 0x77D598
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x75FAA8
        elif rom.region == Region.E:
            return 0x773948
        elif rom.region == Region.J:
            return 0x75FBB8
        elif rom.region == Region.C:
            return 0x79ECA0

    raise ValueError("Rom has unknown game loaded.")


def area_connections(rom: Rom) -> int:
    """Returns the address of the area connections list."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x3C8B90
        elif rom.region == Region.E:
            return 0x3C91EC
        elif rom.region == Region.J:
            return 0x3CB158
        elif rom.region == Region.C:
            return 0x3CB19C
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x360274
        elif rom.region == Region.E:
            return 0x360F00
        elif rom.region == Region.J:
            return 0x3602D0
        elif rom.region == Region.C:
            return 0x379A60

    raise ValueError("Rom has unknown game loaded.")


def area_connections_count(rom: Rom) -> int:
    """Returns the number of area connections in the game. Excludes the final entry of FFs."""
    if rom.game == Game.MF:
        return 0x22
    elif rom.game == Game.ZM:
        return 0x19

    raise ValueError("Rom has unknown game loaded.")


def hatch_lock_events(rom: Rom) -> int:
    """Returns the address of the hatch lock events."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x3C8A5C
        elif rom.region == Region.E:
            return 0x3C90B8
        elif rom.region == Region.J:
            return 0x3CB024
        elif rom.region == Region.C:
            return 0x3CB068
    raise ValueError(rom.game)


def hatch_lock_event_count(rom: Rom) -> int:
    """Returns the number of hatch lock events in the game."""
    if rom.game == Game.MF:
        # Non-vanilla (original: 0x4B)
        return 0xF
    raise ValueError(rom.game)


def starting_equipment(rom: Rom) -> int:
    """Returns the address of the starting equipment data."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x28D2AC
        elif rom.region == Region.E:
            return 0x28D908
        elif rom.region == Region.J:
            return 0x28F5B4
        elif rom.region == Region.C:
            return 0x28F5F8
    raise NotImplementedError(rom.game)


def anim_palette_entries(rom: Rom) -> int:
    """Returns the address of the animated palette entries."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x3E3764
        elif rom.region == Region.E:
            return 0x3E3DC0
        elif rom.region == Region.J:
            return 0x3E5D38
        elif rom.region == Region.C:
            return 0x3E5D7C
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x35FBFC
        elif rom.region == Region.E:
            return 0x360888
        elif rom.region == Region.J:
            return 0x35FC58
        elif rom.region == Region.C:
            return 0x3793E8

    raise ValueError("Rom has unknown game loaded.")


def anim_palette_count(rom: Rom) -> int:
    """Returns the number of animated palettes in the game."""
    if rom.game == Game.MF:
        if rom.region == Region.U or rom.region == Region.E:
            return 0x21
        elif rom.region == Region.J or rom.region == Region.C:
            return 0x22
    elif rom.game == Game.ZM:
        return 0x12
    raise ValueError(rom.game, rom.region)


def sprite_vram_sizes(rom: Rom) -> int:
    """Returns the address of the sprite VRAM sizes."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x2E4A50
        elif rom.region == Region.E:
            return 0x2E50AC
        elif rom.region == Region.J:
            return 0x2E6D58
        elif rom.region == Region.C:
            return 0x2E6D9C
    raise ValueError(rom.game)


def sprite_graphics_ptrs(rom: Rom) -> int:
    """Returns the address of the sprite graphics pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79A5D8
        elif rom.region == Region.E:
            return 0x79AE0C
        elif rom.region == Region.J:
            return 0x7ECC88
        elif rom.region == Region.C:
            return 0x77C2DC
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x75EBF8
        elif rom.region == Region.E:
            return 0x772A98
        elif rom.region == Region.J:
            return 0x75ED08
        elif rom.region == Region.C:
            return 0x79DDF0
    raise ValueError(rom.game, rom.region)


def sprite_palette_ptrs(rom: Rom) -> int:
    """Returns the address of the sprite palette pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79A8D4
        elif rom.region == Region.E:
            return 0x79B108
        elif rom.region == Region.J:
            return 0x7ECF84
        elif rom.region == Region.C:
            return 0x77C5D8
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x75EEF0
        elif rom.region == Region.E:
            return 0x772D90
        elif rom.region == Region.J:
            return 0x75F000
        elif rom.region == Region.C:
            return 0x79E0E8
    raise ValueError(rom.game, rom.region)


def sprite_count(rom: Rom) -> int:
    """Returns the number of sprites in the game."""
    if rom.game == Game.MF:
        return 0xCF
    elif rom.game == Game.ZM:
        return 0xCE
    raise ValueError(rom.game)


def spriteset_ptrs(rom: Rom) -> int:
    """Returns the address of the spriteset pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79ADD8
        elif rom.region == Region.E:
            return 0x79B60C
        elif rom.region == Region.J:
            return 0x7ED488
        elif rom.region == Region.C:
            return 0x77CADC
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x75F31C
        elif rom.region == Region.E:
            return 0x7731BC
        elif rom.region == Region.J:
            return 0x75F42C
        elif rom.region == Region.C:
            return 0x79E514
    raise ValueError(rom.game, rom.region)


def spriteset_count(rom: Rom) -> int:
    """Returns the number of spritesets in the game."""
    if rom.game == Game.MF:
        return 0x82
    elif rom.game == Game.ZM:
        return 0x72
    raise ValueError(rom.game)


def samus_palettes(rom: Rom) -> list[tuple[int, int]]:
    """Returns a list of (address, row count) pairs for all of Samus's palettes."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return [(0x28DD7C, 0x5E), (0x28EAFC, 0x70), (0x565D48, 3)]
        elif rom.region == Region.E:
            return [(0x28E3D8, 0x5E), (0x28F158, 0x70), (0x5663A4, 3)]
        elif rom.region == Region.J:
            return [(0x290084, 0x5E), (0x290E04, 0x70), (0x568424, 3)]
        elif rom.region == Region.C:
            return [(0x2900C8, 0x5E), (0x290E48, 0x70), (0x56CC68, 3)]
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return [(0x2376A8, 0xA3)]
        elif rom.region == Region.E:
            return [(0x238334, 0xA3)]
        elif rom.region == Region.J:
            return [(0x237704, 0xA3)]
        elif rom.region == Region.C:
            return [(0x250E94, 0xA3)]
    raise ValueError(rom.game, rom.region)


def helmet_cursor_palettes(rom: Rom) -> list[tuple[int, int]]:
    """
    Returns a list of (address, row count) pairs for Samus's helmet as a cursor
    (file select and game over)
    """
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return [(0x740E08, 1), (0x740EA8, 2), (0x73C544, 1), (0x73C584, 2)]
        elif rom.region == Region.E:
            return [(0x741618, 1), (0x7416B8, 2), (0x73CD54, 1), (0x73CD94, 2)]
        elif rom.region == Region.J:
            return [(0x73FCDC, 1), (0x73FD7C, 2), (0x73C030, 1), (0x73C070, 2)]
        elif rom.region == Region.C:
            return [(0x6CE360, 1), (0x6CE400, 2), (0x6CA8F8, 1), (0x6CA938, 2)]
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return [(0x454938, 1), (0x4549B8, 1)]
        elif rom.region == Region.E:
            return [(0x4603F8, 1), (0x460478, 1)]
        elif rom.region == Region.J:
            return [(0x454994, 1), (0x454A14, 1)]
        elif rom.region == Region.C:
            return [(0x4768FC, 1), (0x47697C, 1)]
    raise ValueError(rom.game, rom.region)


def beam_palettes(rom: Rom) -> list[tuple[int, int]]:
    """Returns a list of (address, row count) pairs for beam palettes."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return [(0x58B464, 6)]
        elif rom.region == Region.E:
            return [(0x58BAC0, 6)]
        elif rom.region == Region.J:
            return [(0x58BBF4, 6)]
        elif rom.region == Region.C:
            return [(0x592578, 6)]
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return [(0x3270E8, 6)]
        elif rom.region == Region.E:
            return [(0x327D74, 6)]
        elif rom.region == Region.J:
            return [(0x327144, 6)]
        elif rom.region == Region.C:
            return [(0x3408D4, 6)]
    raise ValueError(rom.game, rom.region)


def sax_palettes(rom: Rom) -> list[tuple[int, int]]:
    """Returns a list of (address, row count) pairs for all of the SA-X's palettes."""
    if rom.game == Game.MF:
        # Normal, Lab, Monster, Extra
        if rom.region == Region.U:
            return [(0x2E7D60, 2), (0x2E91D8, 2), (0x38CFB4, 8), (0x2B4368, 5)]
        elif rom.region == Region.E:
            return [(0x2E83BC, 2), (0x2E9834, 2), (0x38D610, 8), (0x2B49C4, 5)]
        elif rom.region == Region.J:
            return [(0x2EA068, 2), (0x2EB4E0, 2), (0x38F2BC, 8), (0x2B6670, 5)]
        elif rom.region == Region.C:
            return [(0x2EA0AC, 2), (0x2EB524, 2), (0x38F300, 8), (0x2B66B4, 5)]
    raise ValueError(rom.game)


def tourian_statues_cutscene_palette(rom: Rom) -> int:
    """Returns the address of the palette used for the Tourian statue cutscenes."""
    if rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x3ED53C
        elif rom.region == Region.E:
            return 0x3EE1C8
        elif rom.region == Region.J:
            return 0x3ED598
        elif rom.region == Region.C:
            return 0x406D28
    raise ValueError(rom.game)


def file_screen_text_ptrs(rom: Rom) -> int:
    """Returns the address of the file screen text pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79EC68
        elif rom.region == Region.E:
            return 0x79F4C4
        elif rom.region == Region.J:
            return 0x7F13FC
    raise ValueError(rom.game)


def character_widths(rom: Rom) -> int:
    """Returns the address of the character widths."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x576234
        elif rom.region == Region.E:
            return 0x576890
        elif rom.region == Region.J:
            return 0x578934
        elif rom.region == Region.C:
            return 0x57D21C
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x40D7B0
        elif rom.region == Region.E:
            return 0x40E5E4
        elif rom.region == Region.J:
            return 0x40D80C
        elif rom.region == Region.C:
            return 0x42F34C
    raise ValueError(rom.game, rom.region)


def navigation_text_ptrs(rom: Rom) -> int:
    """Returns the address of the navigation text pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79C0F0
        elif rom.region == Region.E:
            return 0x79C924
        elif rom.region == Region.J:
            return 0x7EE7A0
        elif rom.region == Region.C:
            return 0x77DDF4
    raise ValueError(rom.game)


def sound_data_entries(rom: Rom) -> int:
    """Returns the address of the sound data entries."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0xA8D3C
        elif rom.region == Region.E:
            return 0xA9398
        elif rom.region == Region.J:
            return 0xAB0A0
        elif rom.region == Region.C:
            return 0xAB0E4
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x8F2C0
        elif rom.region == Region.E:
            return 0x8FF4C
        elif rom.region == Region.J:
            return 0x8F31C
        elif rom.region == Region.C:
            return 0xA8AAC
    raise ValueError(rom.game, rom.region)


def sound_count(rom: Rom) -> int:
    """Returns the number of sounds in the game."""
    if rom.game == Game.MF:
        return 0x2E9
    elif rom.game == Game.ZM:
        return 0x2C4
    raise ValueError(rom.game)


def minimap_ptrs(rom: Rom) -> int:
    """Returns the address of the minimap data pointers."""
    if rom.game == Game.MF:
        if rom.region == Region.U:
            return 0x79BE5C
        elif rom.region == Region.E:
            return 0x79C690
        elif rom.region == Region.J:
            return 0x7EE50C
        elif rom.region == Region.C:
            return 0x77DB60
    elif rom.game == Game.ZM:
        if rom.region == Region.U:
            return 0x7601EC
        elif rom.region == Region.E:
            return 0x77408C
        elif rom.region == Region.J:
            return 0x7602FC
        elif rom.region == Region.C:
            return 0x79F3EC
    raise ValueError(rom.game, rom.region)


def minimap_count(rom: Rom) -> int:
    """Returns the number of minimaps in the game."""
    if rom.game == Game.MF:
        return 11
    elif rom.game == Game.ZM:
        return 11
    raise ValueError(rom.game)



from mars_patcher.auto_generated_types import MarsschemaTankincrements
from mars_patcher.constants.reserved_space import ReservedConstants
from mars_patcher.locations import (
    ItemMessages,
    ItemMessagesKind,
    ItemSprite,
    ItemType,
    LocationSettings,
)
from mars_patcher.rom import Rom
from mars_patcher.room_entry import RoomEntry
from mars_patcher.text import Language, MessageType, encode_text
from mars_patcher.tileset import Tileset

MINOR_LOCS_TABLE_ADDR = ReservedConstants.MINOR_LOCS_TABLE_ADDR
MINOR_LOCS_ARRAY_ADDR = ReservedConstants.MINOR_LOCS_ARRAY_ADDR
MINOR_LOC_SIZE = 0x10
MAJOR_LOCS_POINTER_ADDR = ReservedConstants.MAJOR_LOCS_POINTER_ADDR
MAJOR_LOC_SIZE = 0x4
TANK_INC_ADDR = ReservedConstants.TANK_INC_ADDR
REQUIRED_METROID_COUNT_ADDR = ReservedConstants.REQUIRED_METROID_COUNT_ADDR
TOTAL_METROID_COUNT_ADDR = ReservedConstants.TOTAL_METROID_COUNT_ADDR
MESSAGE_TABLE_LOOKUP_ADDR = ReservedConstants.MESSAGE_TABLE_LOOKUP_ADDR
FIRST_CUSTOM_MESSAGE_ID = ReservedConstants.FIRST_CUSTOM_MESSAGE_ID
AUTO_MESSAGE_ID = 0xFF

TANK_CLIP = (0x62, 0x63, 0x68)
HIDDEN_TANK_CLIP = (0x64, 0x65, 0x69)
TANK_BG1_START = 0x40
TANK_TILE = (0x50, 0x54, 0x58)


class ItemPatcher:
    """Class for writing item assignments to a ROM."""

    def __init__(self, rom: Rom, settings: LocationSettings):
        self.rom = rom
        self.settings = settings

    def _binary_search_rooms_array(self, start_address: int, room: int) -> int:
        """Returns either the address of the room, or -1 if the room was not found."""
        low_end = 0
        high_end = 16
        while low_end < high_end:
            middle = (low_end + high_end) // 2
            read_value = self.rom.read_8(start_address + middle)
            if read_value < room:
                low_end = middle + 1
            elif read_value > room:
                high_end = middle
            else:
                return start_address + middle

        return -1

    # TODO: Use separate classes for handling tilesets and backgrounds
    def write_items(self) -> None:
        rom = self.rom
        custom_message_id = FIRST_CUSTOM_MESSAGE_ID
        message_table_addrs: dict[Language, int] = {}
        for lang in Language:
            message_table_addrs[lang] = rom.read_ptr(MESSAGE_TABLE_LOOKUP_ADDR + lang.value * 4)
        # Handle minor locations
        minor_locs = self.settings.minor_locs
        MINOR_LOCS_ARRAY = rom.read_ptr(MINOR_LOCS_ARRAY_ADDR)
        prev_area_room = (-1, -1)
        room_tank_count = 0
        total_metroids = 0
        item_messages_to_custom_id: dict[ItemMessages, int] = {}
        for min_loc in minor_locs:
            if min_loc.new_item == ItemType.INFANT_METROID:
                total_metroids += 1

            # Update room tank count
            area_room = (min_loc.area, min_loc.room)
            if area_room == prev_area_room:
                room_tank_count += 1
            else:
                room_tank_count = 1
                prev_area_room = area_room
            tank_slot = room_tank_count - 1

            # Overwrite clipdata
            room = RoomEntry(rom, min_loc.area, min_loc.room)
            val = HIDDEN_TANK_CLIP[tank_slot] if min_loc.hidden else TANK_CLIP[tank_slot]
            with room.load_clip() as clip:
                clip.set_block_value(min_loc.block_x, min_loc.block_y, val)
            # Overwrite BG1 if not hidden
            if not min_loc.hidden:
                # Get tilemap
                tileset = Tileset(rom, room.tileset())
                addr = tileset.rle_tilemap_addr()
                # Find tank in tilemap
                addr += 2 + (TANK_BG1_START * 8)
                tile = TANK_TILE[tank_slot]
                idx = next(i for i in range(16) if rom.read_8(addr + i * 8) == tile)
                val = TANK_BG1_START + idx
                with room.load_bg1() as bg1:
                    bg1.set_block_value(min_loc.block_x, min_loc.block_y, val)

            # Write to minors array
            # Assembly has:
            # - A list that contains pointers to below area array
            # - An array with 16 elements per each area, that contains
            #   sorted internal room ids which, contain minor items
            # - An array right after that contains the index where this room starts in
            #   the big item array
            # - A big array of all items and their attributes.
            area_addr = rom.read_ptr(MINOR_LOCS_TABLE_ADDR) + (min_loc.area * 4)
            rooms_list_addr = rom.read_ptr(area_addr)
            room_entry_addr = self._binary_search_rooms_array(rooms_list_addr, min_loc.room)
            assert room_entry_addr != -1
            room_entry_index = rom.read_8(room_entry_addr + 16)

            found_item = False
            item_index = -1
            item_addr = -1
            while not found_item:
                item_index += 1
                item_addr = MINOR_LOCS_ARRAY + ((room_entry_index + item_index) * MINOR_LOC_SIZE)
                read_area = rom.read_8(item_addr)
                read_room = rom.read_8(item_addr + 1)
                read_block_x = rom.read_8(item_addr + 3)
                read_block_y = rom.read_8(item_addr + 4)

                assert read_area == min_loc.area, (
                    f"area was '{read_area}', but was expected to be {min_loc.area}"
                )
                assert read_room == min_loc.room, (
                    f"room was '{read_room}', but was expected to be {min_loc.room}"
                )
                found_item = (read_block_x == min_loc.block_x) and (read_block_y == min_loc.block_y)

            assert item_addr != -1

            if min_loc.new_item != ItemType.UNDEFINED:
                rom.write_8(item_addr + 5, min_loc.new_item.value)
                if min_loc.item_sprite != ItemSprite.UNCHANGED:
                    rom.write_8(item_addr + 6, min_loc.item_sprite.value)
            # Handle item messages
            if min_loc.item_messages is not None:
                messages = min_loc.item_messages
                # If the kind is Custom Message, write the message to free space, and set the ID
                if messages.kind == ItemMessagesKind.CUSTOM_MESSAGE:
                    # If we already encountered the message before, write the existing message id.
                    if messages in item_messages_to_custom_id:
                        rom.write_8(item_addr + 7, item_messages_to_custom_id[messages])
                    else:
                        self.write_custom_message(
                            custom_message_id,
                            message_table_addrs,
                            item_addr,
                            min_loc.item_messages,
                            False,
                        )
                        item_messages_to_custom_id[messages] = custom_message_id
                        custom_message_id += 1
                # If the kind is Message ID, write that ID
                else:
                    rom.write_8(item_addr + 7, messages.message_id)
            # Write item jingle
            rom.write_8(item_addr + 8, min_loc.item_jingle.value)
        # Handle major locations
        for maj_loc in self.settings.major_locs:
            # Write to majors table
            if maj_loc.new_item != ItemType.UNDEFINED:
                if maj_loc.new_item == ItemType.INFANT_METROID:
                    total_metroids += 1
                addr = rom.read_ptr(MAJOR_LOCS_POINTER_ADDR) + (
                    maj_loc.major_src.value * MAJOR_LOC_SIZE
                )
                rom.write_8(addr, maj_loc.new_item.value)
                # Handle item messages
                if maj_loc.item_messages is not None:
                    messages = maj_loc.item_messages
                    # If the kind is Custom Message, write the message to free space, and set the ID
                    if messages.kind == ItemMessagesKind.CUSTOM_MESSAGE:
                        # If we already encountered the message before, write the existing id.
                        if messages in item_messages_to_custom_id:
                            rom.write_8(addr + 1, item_messages_to_custom_id[messages])
                        else:
                            self.write_custom_message(
                                custom_message_id,
                                message_table_addrs,
                                addr,
                                maj_loc.item_messages,
                                True,
                            )
                            item_messages_to_custom_id[messages] = custom_message_id
                            custom_message_id += 1
                    # If the kind is Message ID, write that ID
                    else:
                        rom.write_8(addr + 1, messages.message_id)
                else:  # Set ID to Auto Message
                    rom.write_8(addr + 1, AUTO_MESSAGE_ID)
                # Write item jingle
                rom.write_8(addr + 2, maj_loc.item_jingle.value)
        # Write total metroid count
        rom.write_8(rom.read_ptr(TOTAL_METROID_COUNT_ADDR), total_metroids)

    def write_custom_message(
        self,
        custom_message_id: int,
        message_table_addrs: dict[Language, int],
        item_addr: int,
        messages: ItemMessages,
        is_major: bool,
    ) -> None:
        assert custom_message_id < 0xFF, (
            f"There can be no more than {0xFF - FIRST_CUSTOM_MESSAGE_ID} custom messages."
        )
        rom = self.rom
        # Minor locations have a different structure than Major locations,
        # so the custom message id is at a different offset.
        rom.write_8(item_addr + (1 if is_major else 7), custom_message_id)
        for lang in Language:
            # English is required to be set - use English as the fallback value
            encoded_text = encode_text(
                rom,
                MessageType.TWO_LINE,
                (
                    messages.item_messages[lang]
                    if lang in messages.item_messages
                    else messages.item_messages[Language.ENGLISH]
                ),
                centered=messages.centered,
            )
            message_addr = rom.reserve_free_space(len(encoded_text) * 2)
            rom.write_ptr(message_table_addrs[lang] + (4 * custom_message_id), message_addr)
            rom.write_16_list(message_addr, encoded_text)


# TODO: Move these?
def set_required_metroid_count(rom: Rom, count: int) -> None:
    rom.write_8(rom.read_ptr(REQUIRED_METROID_COUNT_ADDR) + 1, count)


def set_tank_increments(rom: Rom, data: MarsschemaTankincrements) -> None:
    rom.write_16(rom.read_ptr(TANK_INC_ADDR), data["MissileTank"])
    rom.write_16(rom.read_ptr(TANK_INC_ADDR) + 2, data["EnergyTank"])
    rom.write_16(rom.read_ptr(TANK_INC_ADDR) + 4, data["PowerBombTank"])
    rom.write_16(rom.read_ptr(TANK_INC_ADDR) + 6, data["MissileData"])
    rom.write_16(rom.read_ptr(TANK_INC_ADDR) + 8, data["PowerBombData"])



import shutil
import tempfile
import urllib.request
from pathlib import Path
from zipfile import ZipFile

import requests

VERSION = "0.8.2"
ASSET_NAME = "Randomizer.Patches.zip"
DESTINATION_ASSEMBLY_PATH = (
    Path(__file__)
    .parent.resolve()
    .joinpath("src", "mars_patcher", "data", "patches", "mf_u", "asm")
)


release_url = f"https://api.github.com/repos/MetroidAdvRandomizerSystem/mars-fusion-asm/releases/tags/{VERSION}"
print(f"Fetching {release_url}")
response = requests.get(release_url).json()


for asset in response["assets"]:
    if asset["name"] != ASSET_NAME:
        continue

    print("Correct release found, initalizing download")
    with tempfile.TemporaryDirectory() as temp_ref:
        temp_dir = Path(temp_ref)
        temp_file = temp_dir.joinpath("cache.zip")
        urllib.request.urlretrieve(asset["browser_download_url"], temp_file)
        with ZipFile(temp_file, "r") as zip_ref:
            zip_ref.extractall(temp_dir)

        for file in list(temp_dir.joinpath("bin").iterdir()):
            print(f"Moving {file.name}")
            shutil.move(file, DESTINATION_ASSEMBLY_PATH.joinpath(file.name))

    break

print("Done.")



# This file is generated. Manual changes will be lost
# fmt: off
# ruff: noqa
# mypy: disable-error-code="misc"
from __future__ import annotations

import typing_extensions as typ


# Definitions
Seed: typ.TypeAlias = typ.Annotated[int, '0 <= value <= 2147483647']
Typeu4: typ.TypeAlias = typ.Annotated[int, '0 <= value <= 15']
Typeu5: typ.TypeAlias = typ.Annotated[int, '0 <= value <= 31']
Typeu8: typ.TypeAlias = typ.Annotated[int, '0 <= value <= 255']
Typeu10: typ.TypeAlias = typ.Annotated[int, '0 <= value <= 1023']
Areaid: typ.TypeAlias = typ.Annotated[int, '0 <= value <= 6']
Areaidkey = typ.Literal[
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6'
]
Minimapidkey = typ.Literal[
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    '10'
]
Sectorid: typ.TypeAlias = typ.Annotated[int, '1 <= value <= 6']
Shortcutsectorlist: typ.TypeAlias = typ.Annotated[list[Sectorid], 'len() == 6']
Huerotation: typ.TypeAlias = typ.Annotated[int, '0 <= value <= 360']
Validsources = typ.Literal[
    'MainDeckData',
    'Arachnus',
    'ChargeCoreX',
    'Level1',
    'TroData',
    'Zazabi',
    'Serris',
    'Level2',
    'PyrData',
    'MegaX',
    'Level3',
    'ArcData1',
    'WideCoreX',
    'ArcData2',
    'Yakuza',
    'Nettori',
    'Nightmare',
    'Level4',
    'AqaData',
    'WaveCoreX',
    'Ridley',
    'Boiler',
    'Animals',
    'AuxiliaryPower'
]
Validitems = typ.Literal[
    'None',
    'Level0',
    'Missiles',
    'MorphBall',
    'ChargeBeam',
    'Level1',
    'Bombs',
    'HiJump',
    'SpeedBooster',
    'Level2',
    'SuperMissiles',
    'VariaSuit',
    'Level3',
    'IceMissiles',
    'WideBeam',
    'PowerBombs',
    'SpaceJump',
    'PlasmaBeam',
    'GravitySuit',
    'Level4',
    'DiffusionMissiles',
    'WaveBeam',
    'ScrewAttack',
    'IceBeam',
    'MissileTank',
    'EnergyTank',
    'PowerBombTank',
    'IceTrap',
    'InfantMetroid'
]
Validitemsprites = typ.Literal[
    'Empty',
    'Missiles',
    'Level0',
    'MorphBall',
    'ChargeBeam',
    'Level1',
    'Bombs',
    'HiJump',
    'SpeedBooster',
    'Level2',
    'SuperMissiles',
    'VariaSuit',
    'Level3',
    'IceMissiles',
    'WideBeam',
    'PowerBombs',
    'SpaceJump',
    'PlasmaBeam',
    'GravitySuit',
    'Level4',
    'DiffusionMissiles',
    'WaveBeam',
    'ScrewAttack',
    'IceBeam',
    'MissileTank',
    'EnergyTank',
    'PowerBombTank',
    'Anonymous',
    'ShinyMissileTank',
    'ShinyPowerBombTank',
    'InfantMetroid'
]
Validabilities = typ.Literal[
    'Missiles',
    'MorphBall',
    'ChargeBeam',
    'Bombs',
    'HiJump',
    'SpeedBooster',
    'SuperMissiles',
    'VariaSuit',
    'IceMissiles',
    'WideBeam',
    'PowerBombs',
    'SpaceJump',
    'PlasmaBeam',
    'GravitySuit',
    'DiffusionMissiles',
    'WaveBeam',
    'ScrewAttack',
    'IceBeam'
]
Validelevatortops = typ.Literal[
    'OperationsDeckTop',
    'MainHubToSector1',
    'MainHubToSector2',
    'MainHubToSector3',
    'MainHubToSector4',
    'MainHubToSector5',
    'MainHubToSector6',
    'MainHubTop',
    'HabitationDeckTop',
    'Sector1ToRestrictedLab'
]
Validelevatorbottoms = typ.Literal[
    'OperationsDeckBottom',
    'MainHubBottom',
    'RestrictedLabToSector1',
    'HabitationDeckBottom',
    'Sector1ToMainHub',
    'Sector2ToMainHub',
    'Sector3ToMainHub',
    'Sector4ToMainHub',
    'Sector5ToMainHub',
    'Sector6ToMainHub'
]
Validlanguages = typ.Literal[
    'JapaneseKanji',
    'JapaneseHiragana',
    'English',
    'German',
    'French',
    'Italian',
    'Spanish'
]
Messagelanguages: typ.TypeAlias = dict[Validlanguages, str]

class Itemmessages(typ.TypedDict, total=False):
    Kind: typ.Required[Itemmessageskind]
    Languages: Messagelanguages
    Centered: bool = True
    MessageID: typ.Annotated[int, '0 <= value <= 56']
    """The Message ID, will display one of the predefined messages in the ROM"""

Itemmessageskind = typ.Literal[
    'CustomMessage',
    'MessageID'
]
Jingle = typ.Literal[
    'Minor',
    'Major'
]

class BlocklayerItem(typ.TypedDict, total=False):
    X: Typeu8
    """The X position in the room that should get edited."""

    Y: Typeu8
    """The Y position in the room that should get edited."""

    Value: Typeu10
    """The value that should be used to edit the room. For backgrounds, this is calculated via `((Row-1) * ColumnsInTileset) + (Column-1)`."""

Blocklayer: typ.TypeAlias = typ.Annotated[list[BlocklayerItem], 'Unique items']
Hintlocks = typ.Literal[
    'OPEN',
    'LOCKED',
    'GREY',
    'BLUE',
    'GREEN',
    'YELLOW',
    'RED'
]

# Schema entries

class MarsschemaLocationsMajorlocationsItem(typ.TypedDict):
    Source: Validsources
    """Valid major locations."""

    Item: Validitems
    """Valid items for shuffling."""

    ItemMessages: typ.NotRequired[Itemmessages]
    Jingle: Jingle

class MarsschemaLocationsMinorlocationsItem(typ.TypedDict):
    Area: Areaid
    """The area ID where this item is located."""

    Room: Typeu8
    """The room ID where this item is located."""

    BlockX: Typeu8
    """The X-coordinate in the room where this item is located."""

    BlockY: Typeu8
    """The Y-coordinate in the room where this item is located."""

    Item: Validitems
    """Valid items for shuffling."""

    ItemSprite: typ.NotRequired[Validitemsprites]
    """Valid graphics for minor location items."""

    ItemMessages: typ.NotRequired[Itemmessages]
    Jingle: Jingle

class MarsschemaLocations(typ.TypedDict):
    """Specifies how the item locations in the game should be changed."""

    MajorLocations: typ.Annotated[list[MarsschemaLocationsMajorlocationsItem], 'len() == 23', 'Unique items']
    """Specifies how the major item locations should be changed. A major item location is a location where an item is obtained by defeating a boss or interacting with a device."""

    MinorLocations: typ.Annotated[list[MarsschemaLocationsMinorlocationsItem], 'len() == 103', 'Unique items']
    """Specifies how the minor item locations should be changed. A minor item location is a location where an item is obtained by touching a tank block."""


class MarsschemaStartinglocation(typ.TypedDict):
    """The location the player should spawn at the start of the game."""

    Area: Areaid
    """The area ID of the starting location."""

    Room: Typeu8
    """The room ID of the starting location."""

    BlockX: Typeu8
    """The X-coordinate in the room where the player should spawn. If the room contains a save station, then this value will not be taken into consideration."""

    BlockY: Typeu8
    """The Y-coordinate in the room where the player should spawn. If the room contains a save station, then this value will not be taken into consideration."""

MarsschemaStartingitemsSecuritylevelsItem = typ.Literal[
    0,
    1,
    2,
    3,
    4
]

class MarsschemaStartingitems(typ.TypedDict, total=False):
    Energy: typ.Annotated[int, '1 <= value <= 2099'] = 99
    """How much energy the player should start with on a new save file."""

    Missiles: typ.Annotated[int, '0 <= value <= 999'] = 10
    """How many missiles the player should start with on a new save file (the amount unlocked by collecting missile data)."""

    PowerBombs: typ.Annotated[int, '0 <= value <= 99'] = 10
    """How many power bombs the player should start with on a new save file (the amount unlocked by collecting power bomb data)."""

    Abilities: typ.Annotated[list[Validabilities], 'Unique items'] = []
    """Which abilities the player should start with on a new save file."""

    SecurityLevels: typ.Annotated[list[MarsschemaStartingitemsSecuritylevelsItem], 'Unique items'] = [0]
    """Which security levels will be unlocked from the start."""

    DownloadedMaps: typ.Annotated[list[Areaid], 'Unique items'] = []
    """Which area maps will be downloaded from the start."""


class MarsschemaTankincrements(typ.TypedDict):
    """How much ammo/health tanks provide when collected."""

    MissileTank: typ.Annotated[int, '-1000 <= value <= 1000'] = 5
    """How much ammo missile tanks provide when collected."""

    EnergyTank: typ.Annotated[int, '-2100 <= value <= 2100'] = 100
    """How much health energy tanks provide when collected."""

    PowerBombTank: typ.Annotated[int, '-100 <= value <= 100'] = 2
    """How much ammo power bomb tanks provide when collected."""

    MissileData: typ.NotRequired[typ.Annotated[int, '0 <= value <= 1000']] = 10
    """How much ammo Missile Launcher Data provides when collected."""

    PowerBombData: typ.NotRequired[typ.Annotated[int, '0 <= value <= 100']] = 10
    """How much ammo Power Bomb Data provides when collected."""


class MarsschemaElevatorconnections(typ.TypedDict):
    """Defines the elevator that each elevator connects to."""

    ElevatorTops: typ.Annotated[dict[Validelevatortops, Validelevatorbottoms], 'len() >= 10']
    """Defines the bottom elevator that each top elevator connects to."""

    ElevatorBottoms: typ.Annotated[dict[Validelevatorbottoms, Validelevatortops], 'len() >= 10']
    """Defines the top elevator that each bottom elevator connects to."""


class MarsschemaSectorshortcuts(typ.TypedDict):
    """Defines the sector that each sector shortcut connects to."""

    LeftAreas: Shortcutsectorlist
    """Destination areas on the left side of sectors."""

    RightAreas: Shortcutsectorlist
    """Destination areas on the right side of sectors"""

MarsschemaDoorlocksItemLocktype = typ.Literal[
    'Open',
    'Level0',
    'Level1',
    'Level2',
    'Level3',
    'Level4',
    'Locked'
]

class MarsschemaDoorlocksItem(typ.TypedDict):
    Area: Areaid
    """The area ID where this door is located."""

    Door: Typeu8
    """The door ID of this door."""

    LockType: MarsschemaDoorlocksItemLocktype
    """The type of cover on the hatch."""

MarsschemaPalettesRandomizeKey = typ.Literal[
    'Tilesets',
    'Enemies',
    'Samus',
    'Beams'
]

@typ.final
class MarsschemaPalettesRandomize(typ.TypedDict, total=False):
    """The range to use for rotating palette hues."""

    HueMin: Huerotation = None
    """The minimum value to use for rotating palette hues. If not specified, the patcher will randomly generate one."""

    HueMax: Huerotation = None
    """The maximum value to use for rotating palette hues. If not specified, the patcher will randomly generate one."""


MarsschemaPalettesColorspace = typ.Literal[
    'HSV',
    'Oklab'
]

@typ.final
class MarsschemaPalettes(typ.TypedDict, total=False):
    """Properties for randomized in-game palettes."""

    Seed: Seed = None
    """A number used to initialize the random number generator for palettes. If not specified, the patcher will randomly generate one."""

    Randomize: typ.Required[dict[MarsschemaPalettesRandomizeKey, MarsschemaPalettesRandomize]]
    """What kind of palettes should be randomized."""

    ColorSpace: MarsschemaPalettesColorspace = 'Oklab'
    """The color space to use for rotating palette hues."""

    Symmetric: bool = True
    """Randomly rotates hues in the positive or negative direction true."""


class MarsschemaNavigationtextNavigationterminals(typ.TypedDict, total=False):
    """Assigns each navigation room a specific text."""

    MainDeckWest: str
    """Specifies what text should appear at the west Navigation Terminal in Main Deck."""

    MainDeckEast: str
    """Specifies what text should appear at the east Navigation Terminal in Main Deck."""

    OperationsDeck: str
    """Specifies what text should appear at the Navigation Terminal in Operations Deck."""

    Sector1Entrance: str
    """Specifies what text should appear at the Navigation Terminal in Sector 1."""

    Sector2Entrance: str
    """Specifies what text should appear at the Navigation Terminal in Sector 2."""

    Sector3Entrance: str
    """Specifies what text should appear at the Navigation Terminal in Sector 3."""

    Sector4Entrance: str
    """Specifies what text should appear at the Navigation Terminal in Sector 4."""

    Sector5Entrance: str
    """Specifies what text should appear at the Navigation Terminal in Sector 5."""

    Sector6Entrance: str
    """Specifies what text should appear at the Navigation Terminal in Sector 6."""

    AuxiliaryPower: str
    """Specifies what text should appear at the Navigation Terminal near the Auxiliary Power Station."""

    RestrictedLabs: str
    """Specifies what text should appear at the Navigation Terminal in the Restricted Labs."""


class MarsschemaNavigationtextShiptext(typ.TypedDict, total=False):
    """Assigns the ship specific text."""

    InitialText: str
    """Specifies what text should appear at the initial ship communication."""

    ConfirmText: str
    """Specifies what text should appear at the ship after confirming 'No' on subsequent ship communications."""


@typ.final
class MarsschemaNavigationtext(typ.TypedDict, total=False):
    """Specifies text for a specific language."""

    NavigationTerminals: MarsschemaNavigationtextNavigationterminals
    """Assigns each navigation room a specific text."""

    ShipText: MarsschemaNavigationtextShiptext
    """Assigns the ship specific text."""



class MarsschemaTitletextItem(typ.TypedDict, total=False):
    Text: typ.Annotated[str, '/^[ -~]{0,30}$/']
    """The ASCII text for this line"""

    LineNum: typ.Annotated[int, '0 <= value <= 14']
MarsschemaCreditstextItemLinetype = typ.Literal[
    'Blank',
    'Blue',
    'Red',
    'White1',
    'White2'
]

class MarsschemaCreditstextItem(typ.TypedDict, total=False):
    LineType: typ.Required[MarsschemaCreditstextItemLinetype]
    """The color and line height of the text (or blank)."""

    Text: typ.Annotated[str, '/^[ -~]{0,34}$/']
    """The ASCII text for this line."""

    BlankLines: Typeu8 = 0
    """Inserts the provided number of blank lines after the text line."""

    Centered: bool = True
    """Centers the text horizontally when true."""

MarsschemaNavstationlocksKey = typ.Literal[
    'MainDeckWest',
    'MainDeckEast',
    'OperationsDeck',
    'RestrictedLabs',
    'AuxiliaryPower',
    'Sector1Entrance',
    'Sector2Entrance',
    'Sector3Entrance',
    'Sector4Entrance',
    'Sector5Entrance',
    'Sector6Entrance'
]


@typ.final
class MarsschemaLeveledits(typ.TypedDict, total=False):
    """Specifies the Room ID."""

    BG1: Blocklayer
    """The BG1 layer that should be edited."""

    BG2: Blocklayer
    """The BG2 layer that should be edited."""

    Clipdata: Blocklayer
    """The Clipdata layer that should be edited."""




class MarsschemaMinimapeditsItem(typ.TypedDict, total=False):
    X: Typeu5
    """The X position in the minimap that should get edited."""

    Y: Typeu5
    """The Y position in the minimap that should get edited."""

    Tile: Typeu10
    """The tile value that should be used to edit the minimap."""

    Palette: Typeu4
    """The palette row to use for the tile."""

    HFlip: bool = False
    """Whether the tile should be horizontally flipped or not."""

    VFlip: bool = False
    """Whether the tile should be vertically flipped or not."""



class MarsschemaRoomnamesItem(typ.TypedDict):
    Area: Areaid
    """The area ID where this room is located."""

    Room: Typeu8 = 3
    """The room ID."""

    Name: typ.Annotated[str, 'len() <= 112']
    """Specifies what text should appear for this room. Two lines are available, with an absolute maximum of 56 characters per line, if all characters used are small. Text will auto-wrap if the next word doesn't fit on the line. If the text is too long, it will be truncated.  Use 
 to force a line break. If not provided, will display 'Unknown Room'."""


class Marsschema(typ.TypedDict, total=False):
    """
    Metroid Fusion patching schema
    
    A json schema describing the input for patching Metroid Fusion via mars_patcher.
    """

    SeedHash: typ.Required[typ.Annotated[str, '/^[0-9A-Z]{8}$/']]
    """A seed hash that will be displayed on the file select screen."""

    Locations: typ.Required[MarsschemaLocations]
    """Specifies how the item locations in the game should be changed."""

    RequiredMetroidCount: typ.Required[typ.Annotated[int, '0 <= value <= 20']]
    """The number of infant Metroids that must be collected to beat the game."""

    StartingLocation: MarsschemaStartinglocation
    """The location the player should spawn at the start of the game."""

    StartingItems: MarsschemaStartingitems = None
    TankIncrements: MarsschemaTankincrements = None
    """How much ammo/health tanks provide when collected."""

    ElevatorConnections: MarsschemaElevatorconnections
    """Defines the elevator that each elevator connects to."""

    SectorShortcuts: MarsschemaSectorshortcuts
    """Defines the sector that each sector shortcut connects to."""

    DoorLocks: list[MarsschemaDoorlocksItem]
    """List of all lockable doors and their lock type."""

    Palettes: MarsschemaPalettes = None
    """Properties for randomized in-game palettes."""

    NavigationText: dict[Validlanguages, MarsschemaNavigationtext] = None
    """Specifies text to be displayed at navigation rooms and the ship."""

    TitleText: list[MarsschemaTitletextItem] = None
    """Lines of ascii text to write to the title screen."""

    CreditsText: list[MarsschemaCreditstextItem]
    """Lines of text to insert into the credits."""

    NavStationLocks: dict[MarsschemaNavstationlocksKey, Hintlocks]
    """Sets the required Security Levels for accessing Navigation Terminals."""

    DisableDemos: bool = False
    """Disables title screen demos when true."""

    SkipDoorTransitions: bool = False
    """Makes all door transitions instant when true."""

    StereoDefault: bool = True
    """Forces stereo sound by default when true."""

    DisableMusic: bool = False
    """Disables all music tracks when true."""

    DisableSoundEffects: bool = False
    """Disables all sound effects when true."""

    MissileLimit: Typeu8 = 3
    """Changes how many missiles can be on-screen at a time. The vanilla game has it set to 2, the randomizer changes it to 3 by default. Zero Mission uses 4."""

    UnexploredMap: bool = False
    """When enabled, starts you with a map where all unexplored items and non-visited tiles have a gray background. This is different from the downloaded map stations where there, the full tile is gray."""

    PowerBombsWithoutBombs: bool = False
    """When enabled, lets you use Power Bombs without needing to collect Bomb Data."""

    AccessibilityPatches: bool = False
    """Whether to apply patches for better accessibility."""

    LevelEdits: dict[Areaidkey, dict[str, MarsschemaLeveledits]]
    """Specifies room edits that should be done. These will be applied last."""

    MinimapEdits: dict[Minimapidkey, list[MarsschemaMinimapeditsItem]]
    """Specifies minimap edits that should be done."""

    HideDoorsOnMinimap: bool = False
    """When enabled, hides doors on the minimap. This is automatically enabled when the 'DoorLocks' field is provided."""

    RoomNames: typ.Annotated[list[MarsschemaRoomnamesItem], 'Unique items']
    """Specifies a name to be displayed when the A Button is pressed on the pause menu."""

    RevealHiddenTiles: bool = False
    """When enabled, reveals normally hidden blocks that are breakable by upgrades. Hidden pickup tanks are not revealed regardless of this setting."""

MarsSchema: typ.TypeAlias = Marsschema



from enum import Enum
from typing import Any, Final


class MajorSource(Enum):
    MAIN_DECK_DATA = 0
    ARACHNUS = 1
    CHARGE_CORE_X = 2
    LEVEL_1 = 3
    TRO_DATA = 4
    ZAZABI = 5
    SERRIS = 6
    LEVEL_2 = 7
    PYR_DATA = 8
    MEGA_X = 9
    LEVEL_3 = 10
    ARC_DATA_1 = 11
    WIDE_CORE_X = 12
    ARC_DATA_2 = 13
    YAKUZA = 14
    NETTORI = 15
    NIGHTMARE = 16
    LEVEL_4 = 17
    AQA_DATA = 18
    WAVE_CORE_X = 19
    RIDLEY = 20
    BOILER = 21
    ANIMALS = 22
    AUXILIARYPOWER = 23


class ItemType(Enum):
    UNDEFINED = -1
    NONE = 0
    LEVEL_0 = 1
    MISSILES = 2
    MORPH_BALL = 3
    CHARGE_BEAM = 4
    LEVEL_1 = 5
    BOMBS = 6
    HI_JUMP = 7
    SPEED_BOOSTER = 8
    LEVEL_2 = 9
    SUPER_MISSILES = 10
    VARIA_SUIT = 11
    LEVEL_3 = 12
    ICE_MISSILES = 13
    WIDE_BEAM = 14
    POWER_BOMBS = 15
    SPACE_JUMP = 16
    PLASMA_BEAM = 17
    GRAVITY_SUIT = 18
    LEVEL_4 = 19
    DIFFUSION_MISSILES = 20
    WAVE_BEAM = 21
    SCREW_ATTACK = 22
    ICE_BEAM = 23
    MISSILE_TANK = 24
    ENERGY_TANK = 25
    POWER_BOMB_TANK = 26
    ICE_TRAP = 27
    INFANT_METROID = 28

    def __le__(self, other: Any) -> bool:
        if isinstance(other, ItemType):
            return self.value <= other.value
        return NotImplemented


class ItemSprite(Enum):
    UNCHANGED = -1
    EMPTY = 0
    LEVEL_0 = 1
    MISSILES = 2
    MORPH_BALL = 3
    CHARGE_BEAM = 4
    LEVEL_1 = 5
    BOMBS = 6
    HI_JUMP = 7
    SPEED_BOOSTER = 8
    LEVEL_2 = 9
    SUPER_MISSILES = 10
    VARIA_SUIT = 11
    LEVEL_3 = 12
    ICE_MISSILES = 13
    WIDE_BEAM = 14
    POWER_BOMBS = 15
    SPACE_JUMP = 16
    PLASMA_BEAM = 17
    GRAVITY_SUIT = 18
    LEVEL_4 = 19
    DIFFUSION_MISSILES = 20
    WAVE_BEAM = 21
    SCREW_ATTACK = 22
    ICE_BEAM = 23
    MISSILE_TANK = 24
    ENERGY_TANK = 25
    POWER_BOMB_TANK = 26
    ANONYMOUS = 27
    SHINY_MISSILE_TANK = 28
    SHINY_POWER_BOMB_TANK = 29
    INFANT_METROID = 30


KEY_MAJOR_LOCS: Final = "MajorLocations"
KEY_MINOR_LOCS: Final = "MinorLocations"
KEY_AREA: Final = "Area"
KEY_ROOM: Final = "Room"
KEY_SOURCE: Final = "Source"
KEY_BLOCK_X: Final = "BlockX"
KEY_BLOCK_Y: Final = "BlockY"
KEY_HIDDEN: Final = "Hidden"
KEY_ORIGINAL: Final = "Original"
KEY_ITEM: Final = "Item"
KEY_ITEM_SPRITE: Final = "ItemSprite"
KEY_ITEM_MESSAGES: Final = "ItemMessages"
KEY_ITEM_MESSAGES_KIND: Final = "Kind"
KEY_LANGUAGES: Final = "Languages"
KEY_CENTERED: Final = "Centered"
KEY_MESSAGE_ID: Final = "MessageID"
KEY_ITEM_JINGLE: Final = "Jingle"


class ItemMessagesKind(Enum):
    CUSTOM_MESSAGE = 0
    MESSAGE_ID = 1


class ItemJingle(Enum):
    MINOR = 0
    MAJOR = 1


JINGLE_ENUMS = {"Minor": ItemJingle.MINOR, "Major": ItemJingle.MAJOR}

SOURCE_ENUMS = {
    "MainDeckData": MajorSource.MAIN_DECK_DATA,
    "Arachnus": MajorSource.ARACHNUS,
    "ChargeCoreX": MajorSource.CHARGE_CORE_X,
    "Level1": MajorSource.LEVEL_1,
    "TroData": MajorSource.TRO_DATA,
    "Zazabi": MajorSource.ZAZABI,
    "Serris": MajorSource.SERRIS,
    "Level2": MajorSource.LEVEL_2,
    "PyrData": MajorSource.PYR_DATA,
    "MegaX": MajorSource.MEGA_X,
    "Level3": MajorSource.LEVEL_3,
    "ArcData1": MajorSource.ARC_DATA_1,
    "WideCoreX": MajorSource.WIDE_CORE_X,
    "ArcData2": MajorSource.ARC_DATA_2,
    "Yakuza": MajorSource.YAKUZA,
    "Nettori": MajorSource.NETTORI,
    "Nightmare": MajorSource.NIGHTMARE,
    "Level4": MajorSource.LEVEL_4,
    "AqaData": MajorSource.AQA_DATA,
    "WaveCoreX": MajorSource.WAVE_CORE_X,
    "Ridley": MajorSource.RIDLEY,
    "Boiler": MajorSource.BOILER,
    "Animals": MajorSource.ANIMALS,
    "AuxiliaryPower": MajorSource.AUXILIARYPOWER,
}

ITEM_ENUMS = {
    "Undefined": ItemType.UNDEFINED,
    "None": ItemType.NONE,
    "Level0": ItemType.LEVEL_0,
    "Missiles": ItemType.MISSILES,
    "MorphBall": ItemType.MORPH_BALL,
    "ChargeBeam": ItemType.CHARGE_BEAM,
    "Level1": ItemType.LEVEL_1,
    "Bombs": ItemType.BOMBS,
    "HiJump": ItemType.HI_JUMP,
    "SpeedBooster": ItemType.SPEED_BOOSTER,
    "Level2": ItemType.LEVEL_2,
    "SuperMissiles": ItemType.SUPER_MISSILES,
    "VariaSuit": ItemType.VARIA_SUIT,
    "Level3": ItemType.LEVEL_3,
    "IceMissiles": ItemType.ICE_MISSILES,
    "WideBeam": ItemType.WIDE_BEAM,
    "PowerBombs": ItemType.POWER_BOMBS,
    "SpaceJump": ItemType.SPACE_JUMP,
    "PlasmaBeam": ItemType.PLASMA_BEAM,
    "GravitySuit": ItemType.GRAVITY_SUIT,
    "Level4": ItemType.LEVEL_4,
    "DiffusionMissiles": ItemType.DIFFUSION_MISSILES,
    "WaveBeam": ItemType.WAVE_BEAM,
    "ScrewAttack": ItemType.SCREW_ATTACK,
    "IceBeam": ItemType.ICE_BEAM,
    "MissileTank": ItemType.MISSILE_TANK,
    "EnergyTank": ItemType.ENERGY_TANK,
    "PowerBombTank": ItemType.POWER_BOMB_TANK,
    "IceTrap": ItemType.ICE_TRAP,
    "InfantMetroid": ItemType.INFANT_METROID,
}

ITEM_SPRITE_ENUMS = {
    "Empty": ItemSprite.EMPTY,
    "Level0": ItemSprite.LEVEL_0,
    "Missiles": ItemSprite.MISSILES,
    "MorphBall": ItemSprite.MORPH_BALL,
    "ChargeBeam": ItemSprite.CHARGE_BEAM,
    "Level1": ItemSprite.LEVEL_1,
    "Bombs": ItemSprite.BOMBS,
    "HiJump": ItemSprite.HI_JUMP,
    "SpeedBooster": ItemSprite.SPEED_BOOSTER,
    "Level2": ItemSprite.LEVEL_2,
    "SuperMissiles": ItemSprite.SUPER_MISSILES,
    "VariaSuit": ItemSprite.VARIA_SUIT,
    "Level3": ItemSprite.LEVEL_3,
    "IceMissiles": ItemSprite.ICE_MISSILES,
    "WideBeam": ItemSprite.WIDE_BEAM,
    "PowerBombs": ItemSprite.POWER_BOMBS,
    "SpaceJump": ItemSprite.SPACE_JUMP,
    "PlasmaBeam": ItemSprite.PLASMA_BEAM,
    "GravitySuit": ItemSprite.GRAVITY_SUIT,
    "Level4": ItemSprite.LEVEL_4,
    "DiffusionMissiles": ItemSprite.DIFFUSION_MISSILES,
    "WaveBeam": ItemSprite.WAVE_BEAM,
    "ScrewAttack": ItemSprite.SCREW_ATTACK,
    "IceBeam": ItemSprite.ICE_BEAM,
    "MissileTank": ItemSprite.MISSILE_TANK,
    "EnergyTank": ItemSprite.ENERGY_TANK,
    "PowerBombTank": ItemSprite.POWER_BOMB_TANK,
    "Anonymous": ItemSprite.ANONYMOUS,
    "ShinyMissileTank": ItemSprite.SHINY_MISSILE_TANK,
    "ShinyPowerBombTank": ItemSprite.SHINY_POWER_BOMB_TANK,
    "InfantMetroid": ItemSprite.INFANT_METROID,
}


# TODO: Consider creating "game_enums.py"
BEAM_FLAGS = {"ChargeBeam": 1, "WideBeam": 2, "PlasmaBeam": 4, "WaveBeam": 8, "IceBeam": 0x10}
MISSILE_BOMB_FLAGS = {
    "Missiles": 1,
    "SuperMissiles": 2,
    "IceMissiles": 4,
    "DiffusionMissiles": 8,
    "Bombs": 0x10,
    "PowerBombs": 0x20,
}
SUIT_MISC_FLAGS = {
    "HiJump": 1,
    "SpeedBooster": 2,
    "SpaceJump": 4,
    "ScrewAttack": 8,
    "VariaSuit": 0x10,
    "GravitySuit": 0x20,
    "MorphBall": 0x40,
    "SaxSuit": 0x80,
}



import math
import random

from mars_patcher.color_spaces import HsvColor, OklabColor, RgbBitSize, RgbColor
from mars_patcher.rom import Rom

HUE_VARIATION_RANGE = 180.0
"""The maximum range that hue can be additionally rotated."""


class SineWave:
    STEP = (2 * math.pi) / 16

    def __init__(self, amplitude: float, frequency: float, phase: float):
        self.amplitude = amplitude
        self.frequency = frequency
        self.phase = phase

    @staticmethod
    def generate(max_range: float) -> "SineWave":
        """
        Generates a random sine wave of the form
            y = amplitude * sin(frequency * x + phase)
        where
            0 <= amplitude <= 1
            1/4 <= frequency <= 1
            x increases in steps of 1/16 of a cycle
            0 <= phase <= 2pi (one cycle)
        """
        assert 0 <= max_range <= 1
        # Prefer amplitudes closer to the max, otherwise the variation is often too subtle
        amplitude = random.uniform(max_range / 2, max_range)
        frequency = random.uniform(0.25, 1)
        phase = random.uniform(0, 2 * math.pi)
        return SineWave(amplitude, frequency, phase)

    def calculate_variation(self, x: int) -> float:
        assert 0 <= x < 16
        return self.amplitude * math.sin(self.frequency * x * self.STEP + self.phase)


class ColorChange:
    def __init__(self, hue_shift: float, hue_var: SineWave | None):
        self.hue_shift = hue_shift
        self.hue_var = hue_var

    def _get_hue_shift(self, index: int) -> float:
        shift = self.hue_shift
        if self.hue_var is not None:
            factor = HUE_VARIATION_RANGE / 2
            shift += self.hue_var.calculate_variation(index) * factor
        return shift

    def change_hsv(self, hsv: HsvColor, index: int) -> HsvColor:
        shift = self._get_hue_shift(index)
        hsv.hue = (hsv.hue + shift) % 360
        return hsv

    def change_oklab(self, lab: OklabColor, index: int) -> OklabColor:
        shift = self._get_hue_shift(index)
        # Convert hue shift to radians
        shift *= math.pi / 180
        lab.shift_hue(shift)
        return lab


class Palette:
    def __init__(self, rows: int, rom: Rom, addr: int):
        assert rows >= 1
        self.colors: list[RgbColor] = []
        for i in range(rows * 16):
            rgb = rom.read_16(addr + i * 2)
            color = RgbColor.from_rgb(rgb, RgbBitSize.Rgb5)
            self.colors.append(color)

    def __getitem__(self, key: int) -> RgbColor:
        return self.colors[key]

    def rows(self) -> int:
        return len(self.colors) // 16

    def byte_data(self) -> bytes:
        arr = bytearray()
        for color in self.colors:
            val = color.rgb_15()
            arr.append(val & 0xFF)
            arr.append(val >> 8)
        return bytes(arr)

    def write(self, rom: Rom, addr: int) -> None:
        data = self.byte_data()
        rom.write_bytes(addr, data)

    def change_colors_hsv(self, change: ColorChange, excluded_rows: set[int]) -> None:
        """Apply a color change using HSV color space."""
        black = RgbColor.black()
        white = RgbColor.white_5()
        for row in range(self.rows()):
            if row in excluded_rows:
                continue
            offset = row * 16
            for i in range(16):
                # Skip black and white
                rgb = self.colors[offset + i]
                if rgb == black or rgb == white:
                    continue
                orig_luma = rgb.luma()
                hsv = change.change_hsv(rgb.hsv(), i)
                rgb = hsv.rgb()
                # Rescale luma
                luma_ratio = orig_luma / rgb.luma()
                rgb.red = min(int(rgb.red * luma_ratio), 255)
                rgb.green = min(int(rgb.green * luma_ratio), 255)
                rgb.blue = min(int(rgb.blue * luma_ratio), 255)
                self.colors[offset + i] = rgb

    def change_colors_oklab(self, change: ColorChange, excluded_rows: set[int]) -> None:
        """Apply a color change using Oklab color space."""
        # Convert shift to radians
        for row in range(self.rows()):
            if row in excluded_rows:
                continue
            offset = row * 16
            for i in range(16):
                rgb = self.colors[offset + i]
                lab = change.change_oklab(rgb.oklab(), i)
                self.colors[offset + i] = lab.rgb()



import logging
from collections import defaultdict
from enum import Enum
from typing import Annotated, TypedDict

from mars_patcher.auto_generated_types import MarsschemaDoorlocksItem
from mars_patcher.common_types import AreaId, AreaRoomPair, RoomId
from mars_patcher.constants.game_data import (
    area_doors_ptrs,
    hatch_lock_event_count,
    hatch_lock_events,
)
from mars_patcher.constants.minimap_tiles import (
    ALL_DOOR_TILE_IDS,
    ALL_DOOR_TILES,
    ColoredDoor,
    Edge,
)
from mars_patcher.minimap import Minimap
from mars_patcher.rom import Rom
from mars_patcher.room_entry import BlockLayer, RoomEntry


class HatchLock(Enum):
    OPEN = 0
    LEVEL_0 = 1
    LEVEL_1 = 2
    LEVEL_2 = 3
    LEVEL_3 = 4
    LEVEL_4 = 5
    LOCKED = 6


HATCH_LOCK_ENUMS = {
    "Open": HatchLock.OPEN,
    "Level0": HatchLock.LEVEL_0,
    "Level1": HatchLock.LEVEL_1,
    "Level2": HatchLock.LEVEL_2,
    "Level3": HatchLock.LEVEL_3,
    "Level4": HatchLock.LEVEL_4,
    "Locked": HatchLock.LOCKED,
}

BG1_VALUES = {
    HatchLock.OPEN: 0x4,
    HatchLock.LEVEL_0: 0x6,
    HatchLock.LEVEL_1: 0x8,
    HatchLock.LEVEL_2: 0xA,
    HatchLock.LEVEL_3: 0xC,
    HatchLock.LEVEL_4: 0xE,
    HatchLock.LOCKED: 0x819A,
}

CLIP_VALUES = {
    HatchLock.OPEN: [0x00, 0x00, 0x00, 0x00, 0x00, 0x00],
    HatchLock.LEVEL_0: [0x30, 0x31, 0x32, 0x33, 0x34, 0x35],
    HatchLock.LEVEL_1: [0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B],
    HatchLock.LEVEL_2: [0x40, 0x41, 0x42, 0x43, 0x44, 0x45],
    HatchLock.LEVEL_3: [0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B],
    HatchLock.LEVEL_4: [0x3C, 0x3D, 0x3E, 0x4C, 0x4D, 0x4E],
    HatchLock.LOCKED: [0x10, 0x10, 0x10, 0x10, 0x10, 0x10],
}

CLIP_TO_HATCH_LOCK: dict[int, HatchLock] = {}
for lock, vals in CLIP_VALUES.items():
    for val in vals:
        CLIP_TO_HATCH_LOCK[val] = lock


EXCLUDED_DOORS = {
    (0, 0xB4),  # Restricted lab escape
}

HatchSlot = Annotated[int, "0 <= value <= 5"]

MinimapLocation = tuple[int, int, RoomId]
"""`(X, Y, RoomId)`"""


class MinimapLockChanges(TypedDict, total=False):
    left: HatchLock
    right: HatchLock


# TODO:
# - Optimize by only loading rooms that contain doors to modify
# - Split into more than one function for readability
def set_door_locks(rom: Rom, data: list[MarsschemaDoorlocksItem]) -> None:
    door_locks = parse_door_lock_data(data)

    # Go through all doors in game in order
    doors_ptrs = area_doors_ptrs(rom)
    loaded_rooms: dict[AreaRoomPair, RoomEntry] = {}

    # (AreaID, RoomID): (BG1, Clipdata)
    loaded_bg1_and_clip: dict[AreaRoomPair, tuple[BlockLayer, BlockLayer]] = {}

    # (AreaID, RoomID): (CappedSlot, CaplessSlot)
    orig_room_hatch_slots: dict[AreaRoomPair, tuple[HatchSlot, HatchSlot]] = {}
    new_room_hatch_slots: dict[AreaRoomPair, tuple[HatchSlot, HatchSlot]] = {}

    hatch_slot_changes: dict[AreaRoomPair, dict[HatchSlot, HatchSlot]] = {}

    def factory() -> dict:
        return defaultdict(dict)

    # AreaID: {(MinimapX, MinimapY, RoomID): {"left" | "right": HatchLock}}
    minimap_changes: dict[AreaId, dict[MinimapLocation, MinimapLockChanges]] = defaultdict(factory)

    for area in range(7):
        area_addr = rom.read_ptr(doors_ptrs + area * 4)
        for door in range(256):
            door_addr = area_addr + door * 0xC
            door_type = rom.read_8(door_addr)

            # Check if at end of list
            if door_type == 0:
                break

            # Skip doors that mage marks as deleted
            room = rom.read_8(door_addr + 1)
            if room == 0xFF:
                continue

            # Skip excluded doors and doors that aren't lockable hatches
            lock = door_locks.get((area, door))
            if (area, door) in EXCLUDED_DOORS or door_type & 0xF != 4:
                # Don't log the error if door is open and JSON says to change to open.
                if lock is not None and not (lock is HatchLock.OPEN and door_type & 0xF == 3):
                    logging.error(f"Area {area} door {door} cannot have its lock changed")
                continue

            # Load room's BG1 and clipdata if not already loaded
            area_room = (area, room)
            room_entry = loaded_rooms.get(area_room)
            if room_entry is None:
                room_entry = RoomEntry(rom, area, room)
                bg1 = room_entry.load_bg1()
                clip = room_entry.load_clip()
                loaded_rooms[area_room] = room_entry
                loaded_bg1_and_clip[area_room] = (bg1, clip)
                orig_room_hatch_slots[area_room] = (0, 5)
                new_room_hatch_slots[area_room] = (0, 5)
                hatch_slot_changes[area_room] = {}
            else:
                _tuple = loaded_bg1_and_clip.get(area_room)
                if _tuple is not None:
                    bg1, clip = _tuple

            # Check x exit distance to get facing direction
            x_exit = rom.read_8(door_addr + 7)
            facing_right = x_exit < 0x80
            dx = 1 if facing_right else -1

            # Get hatch position
            hatch_x = rom.read_8(door_addr + 2) + dx
            hatch_y = rom.read_8(door_addr + 4)

            # Get original hatch slot number
            capped_slot, capless_slot = orig_room_hatch_slots[area_room]
            clip_val = clip.get_block_value(hatch_x, hatch_y)
            orig_has_cap = clip_val != 0
            if orig_has_cap:
                # Has cap
                orig_hatch_slot = capped_slot
                capped_slot += 1
            else:
                # Capless
                orig_hatch_slot = capless_slot
                capless_slot -= 1
            orig_room_hatch_slots[area_room] = (capped_slot, capless_slot)

            # Get new hatch slot number
            capped_slot, capless_slot = new_room_hatch_slots[area_room]
            if lock == HatchLock.LOCKED:
                new_hatch_slot = orig_hatch_slot
                # Mark door as deleted
                rom.write_8(door_addr + 1, 0xFF)
            elif (lock is None and orig_has_cap) or (lock is not None and lock != HatchLock.OPEN):
                # Has cap
                new_hatch_slot = capped_slot
                capped_slot += 1
            else:
                # Capless
                new_hatch_slot = capless_slot
                capless_slot -= 1
            new_room_hatch_slots[area_room] = (capped_slot, capless_slot)

            if new_hatch_slot != orig_hatch_slot:
                hatch_slot_changes[area_room][orig_hatch_slot] = new_hatch_slot

            # Map tiles
            if lock is not None:
                screen_offset_x = (hatch_x - 2) // 15
                screen_offset_y = (hatch_y - 2) // 10

                minimap_x = room_entry.map_x + screen_offset_x
                minimap_y = room_entry.map_y + screen_offset_y

                minimap_areas = [area]
                if area == 0:
                    minimap_areas = [0, 9]  # Main deck seemingly has two maps?
                for minimap_area in minimap_areas:
                    map_tile = minimap_changes[minimap_area][minimap_x, minimap_y, room]
                    if facing_right:
                        map_tile["left"] = lock
                    else:
                        map_tile["right"] = lock

            # Overwrite BG1 and clipdata
            if lock is None:
                # Even if a hatch's lock hasn't changed, its slot may have changed
                lock = CLIP_TO_HATCH_LOCK.get(clip_val)
                if lock is None:
                    continue

            bg1_val = BG1_VALUES[lock]
            if facing_right:
                bg1_val += 1

            clip_val = CLIP_VALUES[lock][new_hatch_slot]

            for y in range(4):
                bg1.set_block_value(hatch_x, hatch_y + y, bg1_val)
                clip.set_block_value(hatch_x, hatch_y + y, clip_val)
                bg1_val += 0x10

    # Write BG1 and clipdata for each room
    for bg1, clip in loaded_bg1_and_clip.values():
        bg1.write()
        clip.write()

    fix_hatch_lock_events(rom, hatch_slot_changes)

    change_minimap_tiles(rom, minimap_changes)


def parse_door_lock_data(data: list[MarsschemaDoorlocksItem]) -> dict[AreaRoomPair, HatchLock]:
    """Returns a dictionary of `(AreaID, RoomID): HatchLock` from the input data."""
    door_locks: dict[AreaRoomPair, HatchLock] = {}
    for entry in data:
        area_door = (entry["Area"], entry["Door"])
        lock = HATCH_LOCK_ENUMS[entry["LockType"]]
        door_locks[area_door] = lock
    return door_locks


def fix_hatch_lock_events(
    rom: Rom, hatch_slot_changes: dict[AreaRoomPair, dict[HatchSlot, HatchSlot]]
) -> None:
    hatch_locks_addr = hatch_lock_events(rom)
    count = hatch_lock_event_count(rom)
    for i in range(count):
        addr = hatch_locks_addr + i * 5
        area = rom.read_8(addr + 1)
        room = rom.read_8(addr + 2) - 1
        changes = hatch_slot_changes.get((area, room))
        # Some rooms no longer have doors in rando
        if changes is None:
            continue
        hatch_flags = rom.read_8(addr + 3)
        new_flags = 0
        remain = (1 << 6) - 1
        for prev_slot, new_slot in changes.items():
            if (1 << prev_slot) & hatch_flags != 0:
                new_flags |= 1 << new_slot
            remain &= ~(1 << new_slot)
        new_flags |= hatch_flags & remain
        rom.write_8(addr + 3, new_flags)


def change_minimap_tiles(
    rom: Rom, minimap_changes: dict[AreaId, dict[MinimapLocation, MinimapLockChanges]]
) -> None:
    MAP_EDGES: dict[HatchLock, Edge | ColoredDoor] = {
        HatchLock.OPEN: Edge.DOOR,
        HatchLock.LEVEL_0: Edge.DOOR,
        HatchLock.LEVEL_1: ColoredDoor.BLUE,
        HatchLock.LEVEL_2: ColoredDoor.GREEN,
        HatchLock.LEVEL_3: ColoredDoor.YELLOW,
        HatchLock.LEVEL_4: ColoredDoor.RED,
        HatchLock.LOCKED: Edge.DOOR,
        # HatchLock.LOCKED: Edge.WALL,
    }

    for area, area_map in minimap_changes.items():
        with Minimap(rom, area) as minimap:
            for (x, y, room), tile_changes in area_map.items():
                tile_id, palette, h_flip, v_flip = minimap.get_tile_value(x, y)

                tile_data = ALL_DOOR_TILES[tile_id]

                # Account for h_flip before changing edges
                left = tile_changes.get("left")
                right = tile_changes.get("right")
                if h_flip:
                    left, right = right, left

                # Replace edges
                edges = tile_data.edges
                if left is not None:
                    edges = edges._replace(left=MAP_EDGES[left])
                if right is not None:
                    edges = edges._replace(right=MAP_EDGES[right])
                og_new_tile_data = tile_data._replace(edges=edges)
                new_tile_data = og_new_tile_data

                def tile_exists() -> bool:
                    return new_tile_data in ALL_DOOR_TILE_IDS

                if new_tile_data.content.can_h_flip and not tile_exists():
                    # Try flipping horizontally
                    new_tile_data = og_new_tile_data.h_flip()
                    if tile_exists():
                        h_flip = not h_flip

                if new_tile_data.content.can_v_flip and not tile_exists():
                    # Try flipping vertically
                    new_tile_data = og_new_tile_data.v_flip()
                    if tile_exists():
                        v_flip = not v_flip

                if (
                    new_tile_data.content.can_h_flip
                    and new_tile_data.content.can_v_flip
                    and not tile_exists()
                ):
                    # Try flipping it both ways
                    new_tile_data = og_new_tile_data.v_flip()
                    new_tile_data = new_tile_data.h_flip()
                    if tile_exists():
                        v_flip = not v_flip
                        h_flip = not h_flip

                if not tile_exists():
                    logging.debug(
                        "Could not edit map tile door icons for "
                        f"area {area} room {room:X}. ({x:X}, {y:X})."
                    )
                    logging.debug(f"  Desired tile: {og_new_tile_data.as_str}")
                    logging.debug("  Falling back to unlocked doors.")

                    # Try replacing with open doors
                    if (left is not None) and tile_data.edges.left.is_door:
                        edges = edges._replace(left=Edge.DOOR)
                    if (right is not None) and tile_data.edges.right.is_door:
                        edges = edges._replace(right=Edge.DOOR)
                    new_tile_data = og_new_tile_data._replace(edges=edges)

                    if tile_exists():
                        logging.debug("  Still no luck. Using vanilla tile.")

                    logging.debug("")

                if tile_exists():
                    minimap.set_tile_value(
                        x,
                        y,
                        ALL_DOOR_TILE_IDS[new_tile_data],
                        palette,
                        h_flip,
                        v_flip,
                    )



from __future__ import annotations

from typing import TYPE_CHECKING

from mars_patcher.compress import comp_rle, decomp_rle
from mars_patcher.constants.game_data import area_room_entry_ptrs

if TYPE_CHECKING:
    from types import TracebackType

    from mars_patcher.rom import Rom


class RoomEntry:
    def __init__(self, rom: Rom, area: int, room: int):
        self.rom = rom
        self.addr = rom.read_ptr(area_room_entry_ptrs(rom) + area * 4) + room * 0x3C

    def bg1_ptr(self) -> int:
        return self.addr + 0xC

    def bg2_ptr(self) -> int:
        return self.addr + 0x10

    def clip_ptr(self) -> int:
        return self.addr + 0x14

    def tileset(self) -> int:
        return self.rom.read_8(self.addr)

    def bg1_addr(self) -> int:
        return self.rom.read_ptr(self.bg1_ptr())

    def bg2_addr(self) -> int:
        return self.rom.read_ptr(self.bg2_ptr())

    def clip_addr(self) -> int:
        return self.rom.read_ptr(self.clip_ptr())

    def default_sprite_layout_addr(self) -> int:
        return self.rom.read_ptr(self.addr + 0x20)

    def default_spriteset(self) -> int:
        return self.rom.read_8(self.addr + 0x24)

    def load_bg1(self) -> BlockLayer:
        return BlockLayer(self.rom, self.bg1_ptr())

    def load_bg2(self) -> BlockLayer:
        return BlockLayer(self.rom, self.bg2_ptr())

    def load_clip(self) -> BlockLayer:
        return BlockLayer(self.rom, self.clip_ptr())

    @property
    def map_x(self) -> int:
        return self.rom.read_8(self.addr + 0x35)

    @property
    def map_y(self) -> int:
        return self.rom.read_8(self.addr + 0x36)


class BlockLayer:
    def __enter__(self) -> BlockLayer:
        # We don't need to do anything
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.write()

    def __init__(self, rom: Rom, ptr: int):
        addr = rom.read_ptr(ptr)
        self.rom = rom
        self.pointer = ptr
        self.width = rom.read_8(addr)
        self.height = rom.read_8(addr + 1)
        self.block_data, self.comp_len = decomp_rle(rom.data, addr + 2)

    def get_block_value(self, x: int, y: int) -> int:
        idx = (y * self.width + x) * 2
        if idx >= len(self.block_data):
            raise IndexError(
                f"Block coordinate ({x}, {y}) is out of bounds!"
                f"Room size: ({self.width}, {self.height})"
            )
        return self.block_data[idx] | self.block_data[idx + 1] << 8

    def set_block_value(self, x: int, y: int, value: int) -> None:
        idx = (y * self.width + x) * 2
        if idx >= len(self.block_data):
            raise IndexError(
                f"Block coordinate ({x}, {y}) is out of bounds!"
                f"Room size: ({self.width}, {self.height})"
            )
        self.block_data[idx] = value & 0xFF
        self.block_data[idx + 1] = value >> 8

    def write(self) -> None:
        comp_data = comp_rle(self.block_data)
        comp_len = len(comp_data)
        if comp_len > self.comp_len:
            # Repoint data
            addr = self.rom.reserve_free_space(comp_len + 2)
            self.rom.write_ptr(self.pointer, addr)
        else:
            addr = self.rom.read_ptr(self.pointer)
        self.rom.write_8(addr, self.width)
        self.rom.write_8(addr + 1, self.height)
        self.rom.write_bytes(addr + 2, comp_data)
        self.comp_len = comp_len



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from frozendict import frozendict

from mars_patcher.constants.items import (
    ITEM_ENUMS,
    ITEM_SPRITE_ENUMS,
    JINGLE_ENUMS,
    KEY_AREA,
    KEY_BLOCK_X,
    KEY_BLOCK_Y,
    KEY_CENTERED,
    KEY_HIDDEN,
    KEY_ITEM,
    KEY_ITEM_JINGLE,
    KEY_ITEM_MESSAGES,
    KEY_ITEM_MESSAGES_KIND,
    KEY_ITEM_SPRITE,
    KEY_LANGUAGES,
    KEY_MAJOR_LOCS,
    KEY_MESSAGE_ID,
    KEY_MINOR_LOCS,
    KEY_ORIGINAL,
    KEY_ROOM,
    KEY_SOURCE,
    SOURCE_ENUMS,
    ItemJingle,
    ItemMessagesKind,
    ItemSprite,
    ItemType,
    MajorSource,
)
from mars_patcher.data import get_data_path
from mars_patcher.text import Language

if TYPE_CHECKING:
    from mars_patcher.auto_generated_types import Itemmessages, MarsschemaLocations


class Location:
    def __init__(
        self, area: int, room: int, orig_item: ItemType, new_item: ItemType = ItemType.UNDEFINED
    ):
        if type(self) is Location:
            raise TypeError()
        self.area = area
        self.room = room
        self.orig_item = orig_item
        self.new_item = new_item

    def __str__(self) -> str:
        item_str = self.orig_item.name
        if self.new_item != ItemType.UNDEFINED:
            item_str += "/" + self.new_item.name
        return f"{self.area},0x{self.room:02X}: {item_str}"


class MajorLocation(Location):
    def __init__(
        self,
        area: int,
        room: int,
        major_src: MajorSource,
        orig_item: ItemType,
        new_item: ItemType = ItemType.UNDEFINED,
        item_messages: ItemMessages | None = None,
        item_jingle: ItemJingle = ItemJingle.MAJOR,
    ):
        super().__init__(area, room, orig_item, new_item)
        self.major_src = major_src
        self.item_messages = item_messages
        self.item_jingle = item_jingle


class MinorLocation(Location):
    def __init__(
        self,
        area: int,
        room: int,
        block_x: int,
        block_y: int,
        hidden: bool,
        orig_item: ItemType,
        new_item: ItemType = ItemType.UNDEFINED,
        item_sprite: ItemSprite = ItemSprite.UNCHANGED,
        item_messages: ItemMessages | None = None,
        item_jingle: ItemJingle = ItemJingle.MINOR,
    ):
        super().__init__(area, room, orig_item, new_item)
        self.block_x = block_x
        self.block_y = block_y
        self.hidden = hidden
        self.item_sprite = item_sprite
        self.item_messages = item_messages
        self.item_jingle = item_jingle


@dataclass(frozen=True)
class ItemMessages:
    kind: ItemMessagesKind
    item_messages: frozendict[Language, str]
    centered: bool
    message_id: int

    LANG_ENUMS: ClassVar[dict[str, Language]] = {
        "JapaneseKanji": Language.JAPANESE_KANJI,
        "JapaneseHiragana": Language.JAPANESE_HIRAGANA,
        "English": Language.ENGLISH,
        "German": Language.GERMAN,
        "French": Language.FRENCH,
        "Italian": Language.ITALIAN,
        "Spanish": Language.SPANISH,
    }

    KIND_ENUMS: ClassVar[dict[str, ItemMessagesKind]] = {
        "CustomMessage": ItemMessagesKind.CUSTOM_MESSAGE,
        "MessageID": ItemMessagesKind.MESSAGE_ID,
    }

    @classmethod
    def from_json(cls, data: Itemmessages) -> ItemMessages:
        item_messages: dict[Language, str] = {}
        centered = True
        kind: ItemMessagesKind = cls.KIND_ENUMS[data[KEY_ITEM_MESSAGES_KIND]]
        message_id = 0
        if kind == ItemMessagesKind.CUSTOM_MESSAGE:
            for lang_name, message in data[KEY_LANGUAGES].items():
                lang = cls.LANG_ENUMS[lang_name]
                item_messages[lang] = message
            centered = data.get(KEY_CENTERED, True)
        else:
            message_id = data[KEY_MESSAGE_ID]

        return cls(kind, frozendict(item_messages), centered, message_id)


class LocationSettings:
    def __init__(self, major_locs: list[MajorLocation], minor_locs: list[MinorLocation]):
        self.major_locs = major_locs
        self.minor_locs = minor_locs

    @classmethod
    def initialize(cls) -> LocationSettings:
        with open(get_data_path("locations.json")) as f:
            data = json.load(f)

        major_locs = []
        for entry in data[KEY_MAJOR_LOCS]:
            major_loc = MajorLocation(
                entry[KEY_AREA],
                entry[KEY_ROOM],
                SOURCE_ENUMS[entry[KEY_SOURCE]],
                ITEM_ENUMS[entry[KEY_ORIGINAL]],
            )
            major_locs.append(major_loc)

        minor_locs = []
        for entry in data[KEY_MINOR_LOCS]:
            minor_loc = MinorLocation(
                entry[KEY_AREA],
                entry[KEY_ROOM],
                entry[KEY_BLOCK_X],
                entry[KEY_BLOCK_Y],
                entry[KEY_HIDDEN],
                ITEM_ENUMS[entry[KEY_ORIGINAL]],
            )
            minor_locs.append(minor_loc)

        return LocationSettings(major_locs, minor_locs)

    def set_assignments(self, data: MarsschemaLocations) -> None:
        for maj_loc_entry in data[KEY_MAJOR_LOCS]:
            # Get source and item
            source = SOURCE_ENUMS[maj_loc_entry[KEY_SOURCE]]
            item = ITEM_ENUMS[maj_loc_entry[KEY_ITEM]]
            # Find location with this source
            maj_loc = next(m for m in self.major_locs if m.major_src == source)
            maj_loc.new_item = item
            if KEY_ITEM_MESSAGES in maj_loc_entry:
                maj_loc.item_messages = ItemMessages.from_json(maj_loc_entry[KEY_ITEM_MESSAGES])
            maj_loc.item_jingle = JINGLE_ENUMS[maj_loc_entry[KEY_ITEM_JINGLE]]
        for min_loc_entry in data[KEY_MINOR_LOCS]:
            # Get area, room, block X, block Y
            area = min_loc_entry[KEY_AREA]
            room = min_loc_entry[KEY_ROOM]
            block_x = min_loc_entry[KEY_BLOCK_X]
            block_y = min_loc_entry[KEY_BLOCK_Y]
            # Find location with this source
            try:
                min_loc = next(
                    m
                    for m in self.minor_locs
                    if m.area == area
                    and m.room == room
                    and m.block_x == block_x
                    and m.block_y == block_y
                )
            except StopIteration:
                raise ValueError(
                    f"Invalid minor location: Area {area}, Room {room}, X {block_x}, Y {block_y}"
                )
            # Set item and item sprite
            item = ITEM_ENUMS[min_loc_entry[KEY_ITEM]]
            min_loc.new_item = item
            if KEY_ITEM_SPRITE in min_loc_entry:
                min_loc.item_sprite = ITEM_SPRITE_ENUMS[min_loc_entry[KEY_ITEM_SPRITE]]
            if KEY_ITEM_MESSAGES in min_loc_entry:
                min_loc.item_messages = ItemMessages.from_json(min_loc_entry[KEY_ITEM_MESSAGES])
            min_loc.item_jingle = JINGLE_ENUMS[min_loc_entry[KEY_ITEM_JINGLE]]



from enum import Enum
from os import PathLike

from mars_patcher.constants.reserved_space import ReservedConstants

BytesLike = bytes | bytearray

SIZE_8MB = 0x800000
ROM_OFFSET = 0x8000000


class Game(Enum):
    """The possible GBA games."""

    MF = 1
    """Metroid Fusion"""
    ZM = 2
    """Zero Mission"""


class Region(Enum):
    """The possible regions."""

    U = 1
    """USA"""
    E = 2
    """European"""
    J = 3
    """Japanese"""
    C = 4
    """Chinese"""


class Rom:
    """
    A class dealing with ROM operations, like loading and saving the ROM, or
    reading and writing bytes.

    Attributes:
        game: An enum indicating the current game that is loaded.
        region: An enum indicating the region of the currently loaded game.
        data: A bytearray containing the data from a loaded game.
        free_space_addr: An integer keeping track of the current known address where free space in
                         the game is contained.
    """

    _title_to_game = {
        # Fusion
        "METROID4USA\0AMTE": {
            "game": Game.MF,
            "region": Region.U,
        },
        "METROID4EUR\0AMTP": {
            "game": Game.MF,
            "region": Region.E,
        },
        "METROID4JPN\0AMTJ": {
            "game": Game.MF,
            "region": Region.J,
        },
        "METFUSIONCHNAMTC": {
            "game": Game.MF,
            "region": Region.C,
        },
        # Zero Mission
        "ZEROMISSIONEBMXE": {
            "game": Game.ZM,
            "region": Region.U,
        },
        "ZEROMISSIONPBMXP": {
            "game": Game.ZM,
            "region": Region.E,
        },
        "ZEROMISSIONJBMXJ": {
            "game": Game.ZM,
            "region": Region.J,
        },
        "ZEROMISSIONCBMXC": {
            "game": Game.ZM,
            "region": Region.C,
        },
    }

    def __init__(self, path: str | PathLike[str]):
        # Read file
        with open(path, "rb") as f:
            self.data = bytearray(f.read())
        # Check length
        if len(self.data) != SIZE_8MB:
            raise ValueError("ROM should be 8MB")
        # Check title and code
        title = self.read_ascii(0xA0, 0x10)
        if title not in self._title_to_game:
            raise ValueError("Not a valid GBA Metroid ROM")

        game = self._title_to_game[title]["game"]
        region = self._title_to_game[title]["region"]
        assert isinstance(game, Game)
        assert isinstance(region, Region)
        self.game = game
        self.region = region

        # For now we only allow MF U
        if self.game == Game.ZM:
            raise ValueError("Not compatible with Metroid Zero Mission")
        if self.region != Region.U:
            raise ValueError("Only compatible with the North American (U) version")
        # Set free space address
        if self.is_mf():
            self.free_space_addr = ReservedConstants.PATCHER_FREE_SPACE_ADDR
        elif self.is_zm():
            raise NotImplementedError()

    def is_mf(self) -> bool:
        """Returns true when the currently loaded game is Metroid Fusion."""
        return self.game == Game.MF

    def is_zm(self) -> bool:
        """Returns true when the currently loaded game is Metroid Zero Mission."""
        return self.game == Game.ZM

    def read_8(self, addr: int) -> int:
        """Reads one byte from the specified address, and returns the read value."""
        return self.data[addr]

    def read_16(self, addr: int) -> int:
        """Reads two bytes from the specified address, and returns the read value."""
        return self.data[addr] | (self.data[addr + 1] << 8)

    def read_32(self, addr: int) -> int:
        """Reads four bytes from the specified address, and returns the read value."""
        return (
            self.data[addr]
            | (self.data[addr + 1] << 8)
            | (self.data[addr + 2] << 16)
            | (self.data[addr + 3] << 24)
        )

    def read_ptr(self, addr: int) -> int:
        """
        Reads a pointer (four bytes) from the specified address, and returns the read address.
        The return value is adjusted to be used right away, by subtracting the offset on where
        the ROM is loaded in the GBA.

        Raises:
            ValueError: If the pointer that was read does not point to the region where
                        the GBA loads ROMs.
        """
        val = self.read_32(addr)
        if val < ROM_OFFSET:
            raise ValueError(f"Invalid pointer {val:X} at {addr:X}")
        return val - ROM_OFFSET

    def read_bytes(self, addr: int, size: int) -> bytearray:
        """
        Reads a specified amount of bytes from a given address, and returns
        the read values as a bytearray.
        """
        end = addr + size
        return self.data[addr:end]

    def read_ascii(self, addr: int, size: int) -> str:
        """
        Reads a specified amount of bytes from a given addres, and returns
        the read values interpreted as an ASCII string
        """
        return self.read_bytes(addr, size).decode("ascii")

    def write_8(self, addr: int, val: int) -> None:
        """Writes a number as a byte to a specified address."""
        self.data[addr] = val & 0xFF

    def write_16(self, addr: int, val: int) -> None:
        """Writes a number as two bytes (short) to a specified address."""
        val &= 0xFFFF
        self.data[addr] = val & 0xFF
        self.data[addr + 1] = val >> 8

    def write_32(self, addr: int, val: int) -> None:
        """Writes a number as four bytes (int) to a specified address."""
        val &= 0xFFFFFFFF
        self.data[addr] = val & 0xFF
        self.data[addr + 1] = (val >> 8) & 0xFF
        self.data[addr + 2] = (val >> 16) & 0xFF
        self.data[addr + 3] = val >> 24

    def write_ptr(self, addr: int, val: int) -> None:
        """
        Writes a pointer to a specified address. The pointer needs to be less than
        the offset where the ROM is loaded in the GBA.
        """
        assert val < ROM_OFFSET, f"Pointer should be less than {ROM_OFFSET:X} but is {val:X}"
        self.write_32(addr, val + ROM_OFFSET)

    def write_bytes(
        self, data_addr: int, vals: BytesLike, val_addr: int = 0, size: int | None = None
    ) -> None:
        """
        Writes a bytes like object either in full or partially to a specified address.

        Args:
            data_addr: Where to write in the ROM data.
            vals: A bytes like object that should get written.
            val_addr: An index from where vals should begin to start writing. 0 by default.
            size: How many bytes should be written to the ROM. If None is specified,
                  then that means the full length of vals. None by default.
        """
        if size is None:
            size = len(vals) - val_addr
        data_end = data_addr + size
        val_end = val_addr + size
        self.data[data_addr:data_end] = vals[val_addr:val_end]

    def write_16_list(self, addr: int, vals: list[int]) -> int:
        """Writes a list of numbers as 16-bit integers. Does not check if the
        values are within 16-bit range. Returns the ending address."""
        for val in vals:
            self.write_16(addr, val)
            addr += 2
        return addr

    def copy_bytes(self, src_addr: int, dst_addr: int, size: int) -> None:
        """Copies a specified amount of bytes from the source address to the destination address."""
        self.write_bytes(dst_addr, self.data, src_addr, size)

    def reserve_free_space(self, size: int) -> int:
        """
        Returns an address that is able to fit in a specified size.
        Alignment is always 4.
        """
        remain = self.free_space_addr % 4
        if remain != 0:
            self.free_space_addr += 4 - remain
        addr = self.free_space_addr
        self.free_space_addr += size
        if self.free_space_addr > ReservedConstants.PATCHER_FREE_SPACE_END:
            raise RuntimeError("Ran out of reserved free space")
        return addr

    def save(self, path: str | PathLike[str]) -> None:
        """Saves the currently loaded data to a specified path."""
        with open(path, "wb") as f:
            f.write(self.data)



import argparse
import json

from mars_patcher.patcher import patch, validate_patch_data


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("rom_path", type=str, help="Path to a GBA ROM file")
    parser.add_argument("out_path", type=str, help="Path to output ROM file")
    parser.add_argument("patch_data_path", type=str, help="Path to patch data json file")
    args = parser.parse_args()

    # Load patch data file and validate
    with open(args.patch_data_path, encoding="utf-8") as f:
        patch_data = json.load(f)

    patch(
        args.rom_path,
        args.out_path,
        validate_patch_data(patch_data),
        lambda message, progress: print(message),
    )



class ReservedConstants:
    """
    These are constants that are in the patches 'Reserved Space';
    things that are intended to be modified by this patcher.
    """

    # These need to be kept in sync with the base patch
    # found somewhere around https://github.com/MetroidAdvRandomizerSystem/MARS-Fusion/blob/main/src/main.s#L45

    # Pointers, offset by language value, that store the message table location
    MESSAGE_TABLE_LOOKUP_ADDR = 0x79CDF4
    FIRST_CUSTOM_MESSAGE_ID = 0x39  # The first 0x38 messages are reserved for standard messages

    PATCHER_FREE_SPACE_ADDR = 0x7D0000
    PATCHER_FREE_SPACE_END = PATCHER_FREE_SPACE_ADDR + 0x20000
    MINOR_LOCS_TABLE_ADDR = 0x7FF000
    MINOR_LOCS_ARRAY_ADDR = 0x7FF004
    MAJOR_LOCS_POINTER_ADDR = 0x7FF008
    TANK_INC_ADDR = 0x7FF00C
    TOTAL_METROID_COUNT_ADDR = 0x7FF010
    REQUIRED_METROID_COUNT_ADDR = 0x7FF010
    STARTING_LOCATION_ADDR = 0x7FF014
    CREDITS_END_DELAY_ADDR = 0x7FF018  # TODO: Is this meant to be changed?
    CREDITS_SCROLL_SPEED_ADDR = 0x7FF018  # + 2 TODO: Ditto
    HINT_SECURITY_LEVELS_ADDR = 0x7FF01C  # TODO: ???
    ENVIRONMENTAL_HARZARD_DAMAGE_ADDR = 0x7FF020  # TODO: Implement this
    MISSILE_LIMIT_ADDR = 0x7FF024
    ROOM_NAMES_TABLE_ADDR = 0x7FF028
    REVEAL_HIDDEN_TILES_ADDR = 0x7FF02C
    TITLESCREEN_TEXT_POINTERS_POINTER_ADDR = 0x7FF030



from enum import Enum


class LineType(Enum):
    BLANK = 0
    BLUE = 1
    RED = 2
    WHITE1 = 3
    WHITE2 = 4
    COPYRIGHT1 = 5
    COPYRIGHT2 = 6
    COPYRIGHT3 = 7
    COPYRIGHT4 = 8
    END = 9


LINE_TYPE_VALS = {
    LineType.BLANK: 0x5,
    LineType.BLUE: 0x0,
    LineType.RED: 0x1,
    LineType.WHITE1: 0x3,
    LineType.WHITE2: 0x2,
    LineType.COPYRIGHT1: 0xA,
    LineType.COPYRIGHT2: 0xB,
    LineType.COPYRIGHT3: 0xC,
    LineType.COPYRIGHT4: 0xD,
    LineType.END: 0x6,
}

LINE_TYPE_HEIGHTS = {
    LineType.BLANK: 1,
    LineType.BLUE: 1,
    LineType.RED: 1,
    LineType.WHITE1: 1,
    LineType.WHITE2: 2,
    LineType.COPYRIGHT1: 1,
    LineType.COPYRIGHT2: 1,
    LineType.COPYRIGHT3: 1,
    LineType.COPYRIGHT4: 1,
    LineType.END: 0,
}

TEXT_LINE_TYPES = {LineType.BLUE, LineType.RED, LineType.WHITE1, LineType.WHITE2}

MARS_CREDITS = [
    (LineType.WHITE2, 4, "MARS Development"),
    (LineType.BLUE, 2, "PROGRAMMING"),
    (LineType.WHITE2, 1, "AntyMew"),
    (LineType.WHITE2, 1, "biospark"),
    (LineType.WHITE2, 1, "cluekitty"),
    (LineType.WHITE2, 1, "Zarakava"),
    (LineType.WHITE2, 1, "Miepee"),
    (LineType.WHITE2, 1, "JeffGainsNGames"),
    (LineType.WHITE2, 3, "duncathan_salt"),
    (LineType.BLUE, 2, "GRAPHICS"),
    (LineType.WHITE2, 1, "jakoliath"),
    (LineType.WHITE2, 1, "AbyssalCreature"),
    (LineType.WHITE2, 1, "bariasu"),
    (LineType.WHITE2, 3, "Raddley Vance"),
    (LineType.BLUE, 2, "MUSIC"),
    (LineType.WHITE2, 3, "JRP"),
    (LineType.BLUE, 2, "SPECIAL THANKS TO"),
    (LineType.WHITE2, 1, "Yohann"),
    (LineType.WHITE2, 2, "Kazuto"),
    (LineType.WHITE2, 2, "And contributions from others"),
]

RDV_CREDITS = [
    (LineType.WHITE2, 4, "Randovania"),
    (LineType.BLUE, 2, "Integration"),
    (LineType.WHITE2, 1, "JeffGainsNGames"),
    (LineType.WHITE2, 1, "Miepee"),
    (LineType.WHITE2, 1, "m4rkersm4sk"),
    (LineType.WHITE2, 3, "cluekitty"),
    (LineType.BLUE, 2, "LOGIC DATABASE"),
    (LineType.WHITE2, 1, "cluekitty"),
    (LineType.WHITE2, 1, "HerculesBenchPress"),
    (LineType.WHITE2, 1, "HyperBola"),
    (LineType.WHITE2, 1, "JaggerG"),
    (LineType.WHITE2, 1, "JeffGainsNGames"),
    (LineType.WHITE2, 1, "Schwartz Ghandi"),
    (LineType.WHITE2, 1, "m4rkersm4sk"),
    (LineType.WHITE2, 1, "Miepee"),
    (LineType.WHITE2, 3, "Sharky"),
]

# Shortened from 462 lines to 308 lines
FUSION_STAFF_LINES = [
    (LineType.WHITE2, 4, "ORIGINAL FUSION STAFF"),
    (LineType.BLUE, 1, "PRODUCER"),
    (LineType.WHITE1, 5, "Izushi Takehiro"),
    (LineType.BLUE, 1, "CHIEF DIRECTOR"),
    (LineType.WHITE1, 5, "Sakamoto Yoshio"),
    (LineType.BLUE, 1, "SYSTEM DIRECTOR"),
    (LineType.WHITE1, 3, "Hosokawa Takehiko"),
    (LineType.BLUE, 1, "SCENARIO & STORY"),
    (LineType.WHITE1, 3, "Sakamoto Yoshio"),
    (LineType.BLUE, 1, "GAME DESIGN"),
    (LineType.WHITE1, 1, "Yamane Tomoyoshi"),
    (LineType.WHITE1, 3, "Hosokawa Takehiko"),
    (LineType.BLUE, 1, "COURSE DESIGN"),
    (LineType.WHITE1, 1, "Ueda Masani"),
    (LineType.WHITE1, 3, "Nakada Ryuichi"),
    (LineType.BLUE, 1, "VISUAL DIRECTOR"),
    (LineType.WHITE1, 3, "Yamane Tomoyoshi"),
    (LineType.BLUE, 1, "SAMUS DESIGN"),
    (LineType.WHITE1, 3, "Sano Shinya"),
    (LineType.BLUE, 1, "SAMUS ORIGINAL DESIGN"),
    (LineType.WHITE1, 3, "Kiyotake Hiroji"),
    (LineType.BLUE, 1, "ENEMY CHARACTER DESIGN"),
    (LineType.WHITE1, 1, "Yamane Tomoyoshi"),
    (LineType.WHITE1, 1, "Hirano Isao"),
    (LineType.WHITE1, 1, "Sano Shinya"),
    (LineType.WHITE1, 3, "Matsushita Tadashi"),
    (LineType.BLUE, 1, "BACKGROUND GRAPHIC DESIGN"),
    (LineType.WHITE1, 1, "Morisawa Takayasu"),
    (LineType.WHITE1, 3, "Miyamoto Fumiko"),
    (LineType.BLUE, 1, "GRAPHIC DESIGN"),
    (LineType.WHITE1, 1, "Momose Hiroshi"),
    (LineType.WHITE1, 3, "Sakiyama Chiharu"),
    (LineType.BLUE, 1, "PROGRAM DIRECTOR"),
    (LineType.WHITE1, 3, "Yamano Katsuya"),
    (LineType.BLUE, 1, "SAMUS PROGRAMMING"),
    (LineType.WHITE1, 3, "Katsuki Yoshinori"),
    (LineType.BLUE, 1, "ENEMY CHARACTER PROGRAMMING"),
    (LineType.WHITE1, 3, "Yamano Katsuya"),
    (LineType.BLUE, 1, "SYSTEM PROGRAMMING"),
    (LineType.WHITE1, 3, "Ozaki Nobuhiro"),
    (LineType.BLUE, 1, "CINEMA SCENE PROGRAMMING"),
    (LineType.WHITE1, 3, "Nakae Eisaku"),
    (LineType.BLUE, 1, "TOOL PROGRAMMING"),
    (LineType.WHITE1, 1, "Harada Takahiro"),
    (LineType.WHITE1, 1, "Fukui Kota"),
    (LineType.WHITE1, 1, "Katsuki Yoshinori"),
    (LineType.WHITE1, 1, "Ozaki Nobuhiro"),
    (LineType.WHITE1, 3, "Momose Hiroshi"),
    (LineType.BLUE, 1, "SOUND DIRECTOR"),
    (LineType.WHITE1, 3, "Yamamoto Kenji"),
    (LineType.BLUE, 1, "MUSIC COMPOSITION"),
    (LineType.WHITE1, 1, "Hamano Minako"),
    (LineType.WHITE1, 3, "Fujiwara Akira"),
    (LineType.BLUE, 1, "SOUND EFFECTS & PROGRAMMING"),
    (LineType.WHITE1, 3, "Fujii Takaya"),
    (LineType.BLUE, 1, "PACKAGE DESIGN"),
    (LineType.WHITE1, 3, "Nakamichi Sachiko"),
    (LineType.BLUE, 1, "MANUAL DESIGN"),
    (LineType.WHITE1, 1, "Nakamichi Sachiko"),
    (LineType.WHITE1, 1, "Itani Kiyomi"),
    (LineType.WHITE1, 3, "Takahashi Yasuko"),
    (LineType.BLUE, 1, "COORDINATION"),
    (LineType.WHITE1, 3, "Harada Takahiro"),
    (LineType.BLUE, 1, "INTERNATIONAL COORDINATION"),
    (LineType.WHITE1, 3, "Yamazaki Sakae"),
    (LineType.BLUE, 1, "DEBUG"),
    (LineType.WHITE1, 6, "SUPER MARIO CLUB"),
    (LineType.BLUE, 1, "SPECIAL THANKS"),
    (LineType.WHITE1, 10, "Matsumura Satoshi"),
    (LineType.RED, 2, "NOA ENGLISH LOCALIZATION"),
    (LineType.BLUE, 1, "LOCALIZATION MANAGEMENT"),
    (LineType.WHITE1, 1, "Jeff Miller"),
    (LineType.WHITE1, 3, "Leslie Swan"),
    (LineType.BLUE, 1, "LOCALIZATION"),
    (LineType.WHITE1, 1, "Scot Ritchey"),
    (LineType.WHITE1, 1, "Nathan Bihldorff"),
    (LineType.WHITE1, 3, "Tim O-Leary"),
    (LineType.BLUE, 1, "ENGINEERING DEBUG"),
    (LineType.WHITE1, 1, "Raychole L'Anett"),
    (LineType.WHITE1, 1, "Robert Crombie"),
    (LineType.WHITE1, 1, "Kirk Buchanan"),
    (LineType.WHITE1, 2, "Yoshinobu Mantani"),
    (LineType.WHITE1, 1, "Scott Callahan"),
    (LineType.WHITE1, 1, "Don Dillinger"),
    (LineType.WHITE1, 3, "Hisako Sakauchi"),
    (LineType.BLUE, 1, "SPECIAL THANKS"),
    (LineType.WHITE1, 1, "Tatsumi Kimishima"),
    (LineType.WHITE1, 1, "Mike Fukuda"),
    (LineType.WHITE1, 13, "Hiroshi Kamada"),
    (LineType.BLUE, 1, "EXECUTIVE PRODUCER"),
    (LineType.WHITE1, 22, "Iwata Satoru"),
    (LineType.COPYRIGHT1, 0, None),
    (LineType.COPYRIGHT2, 0, None),
    (LineType.COPYRIGHT3, 0, None),
    (LineType.COPYRIGHT4, 9, None),
    (LineType.END, 0, None),
]



from mars_patcher.auto_generated_types import Areaid, MarsschemaRoomnamesItem, Typeu8
from mars_patcher.constants.reserved_space import ReservedConstants
from mars_patcher.rom import Rom
from mars_patcher.text import MessageType, encode_text

ROOM_NAMES_TABLE_ADDR = ReservedConstants.ROOM_NAMES_TABLE_ADDR


# Write Room Names to ROM
# Assembly has:
# - A list that contains pointers to area room names
# - Area Room names are indexed by room id. This means some entries
#   are never used, but this allows for easy lookup
def write_room_names(rom: Rom, data: list[MarsschemaRoomnamesItem]) -> None:
    seen_rooms: set[tuple[Areaid, Typeu8]] = set()
    for room_name_entry in data:
        area_id = room_name_entry["Area"]
        room_id = room_name_entry["Room"]
        room_name = room_name_entry["Name"]

        # Check that the room wasn't already set
        assert (area_id, room_id) not in seen_rooms, "Duplicate room name provided."
        seen_rooms.add((area_id, room_id))

        # Find room name table by indexing at ROOM_NAMES_TABLE_ADDR
        area_addr = rom.read_ptr(ROOM_NAMES_TABLE_ADDR) + (area_id * 4)
        area_room_name_addr = rom.read_ptr(area_addr)

        # Find specific room by indexing by the room_id
        room_name_addr = area_room_name_addr + (room_id * 4)

        encoded_text = encode_text(rom, MessageType.TWO_LINE, room_name)
        message_addr = rom.reserve_free_space(len(encoded_text) * 2)
        rom.write_ptr(room_name_addr, message_addr)
        rom.write_16_list(message_addr, encoded_text)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from mars_patcher.constants.game_data import navigation_text_ptrs
from mars_patcher.constants.reserved_space import ReservedConstants
from mars_patcher.rom import Rom
from mars_patcher.text import Language, MessageType, encode_text

if TYPE_CHECKING:
    from mars_patcher.auto_generated_types import Hintlocks, MarsschemaNavstationlocksKey
    from mars_patcher.rom import Rom


class NavRoom(Enum):
    MAIN_DECK_WEST = 1
    MAIN_DECK_EAST = 2
    OPERATIONS_DECK = 3
    SECTOR1_ENTRANCE = 4
    SECTOR5_ENTRANCE = 5
    SECTOR2_ENTRANCE = 6
    SECTOR4_ENTRANCE = 7
    SECTOR3_ENTRANCE = 8
    SECTOR6_ENTRANCE = 9
    AUXILIARY_POWER = 10
    RESTRICTED_LABS = 11


# Later down the line, when/if Nav terminals don't have their
# confirm text patched out, combine these two.
class ShipText(Enum):
    INITIAL_TEXT = 0
    CONFIRM_TEXT = 1


class NavStationLockType(Enum):
    OPEN = 0xFF
    LOCKED = 0x05
    GREY = 0x00
    BLUE = 0x01
    GREEN = 0x02
    YELLOW = 0x03
    RED = 0x04


class NavigationText:
    LANG_ENUMS = {
        "JapaneseKanji": Language.JAPANESE_KANJI,
        "JapaneseHiragana": Language.JAPANESE_HIRAGANA,
        "English": Language.ENGLISH,
        "German": Language.GERMAN,
        "French": Language.FRENCH,
        "Italian": Language.ITALIAN,
        "Spanish": Language.SPANISH,
    }

    NAV_ROOM_ENUMS: dict[MarsschemaNavstationlocksKey, NavRoom] = {
        "MainDeckWest": NavRoom.MAIN_DECK_WEST,
        "MainDeckEast": NavRoom.MAIN_DECK_EAST,
        "OperationsDeck": NavRoom.OPERATIONS_DECK,
        "Sector1Entrance": NavRoom.SECTOR1_ENTRANCE,
        "Sector2Entrance": NavRoom.SECTOR2_ENTRANCE,
        "Sector3Entrance": NavRoom.SECTOR3_ENTRANCE,
        "Sector4Entrance": NavRoom.SECTOR4_ENTRANCE,
        "Sector5Entrance": NavRoom.SECTOR5_ENTRANCE,
        "Sector6Entrance": NavRoom.SECTOR6_ENTRANCE,
        "AuxiliaryPower": NavRoom.AUXILIARY_POWER,
        "RestrictedLabs": NavRoom.RESTRICTED_LABS,
    }

    GAME_START_CHAR = "[GAME_START]"
    INITIAL_TEXT_KEY = "InitialText"
    NAV_TERMINALS_KEY = "NavigationTerminals"
    SHIP_TEXT_KEY = "ShipText"

    INFO_TEXT_ENUMS = {
        INITIAL_TEXT_KEY: ShipText.INITIAL_TEXT,
        "ConfirmText": ShipText.CONFIRM_TEXT,
    }

    def __init__(self, navigation_text: dict[Language, dict[str, dict[Enum, str]]]):
        self.navigation_text = navigation_text

    @classmethod
    def from_json(cls, data: dict) -> NavigationText:
        navigation_text: dict[Language, dict[str, dict[Enum, str]]] = {}
        for lang, lang_text in data.items():
            lang = cls.LANG_ENUMS[lang]
            navigation_text[lang] = {
                cls.NAV_TERMINALS_KEY: {
                    cls.NAV_ROOM_ENUMS[k]: v for k, v in lang_text[cls.NAV_TERMINALS_KEY].items()
                },
                cls.SHIP_TEXT_KEY: {
                    # Make sure initial text string starts with [GAME_START]
                    cls.INFO_TEXT_ENUMS[k]: cls.GAME_START_CHAR + v
                    if k == cls.INITIAL_TEXT_KEY and not v.startswith(cls.GAME_START_CHAR)
                    else v
                    for k, v in lang_text[cls.SHIP_TEXT_KEY].items()
                },
            }
        return cls(navigation_text)

    def write(self, rom: Rom) -> None:
        for lang, lang_texts in self.navigation_text.items():
            base_text_address = rom.read_ptr(navigation_text_ptrs(rom) + lang.value * 4)

            # Info Text
            for info_place, text in lang_texts["ShipText"].items():
                encoded_text = encode_text(rom, MessageType.CONTINUOUS, text)
                text_addr = rom.reserve_free_space(len(encoded_text) * 2)
                rom.write_ptr(base_text_address + info_place.value * 4, text_addr)
                rom.write_ptr(base_text_address + info_place.value * 4 + 4, text_addr)
                rom.write_16_list(text_addr, encoded_text)

            # Navigation Text
            for nav_room, text in lang_texts["NavigationTerminals"].items():
                encoded_text = encode_text(rom, MessageType.CONTINUOUS, text)
                text_addr = rom.reserve_free_space(len(encoded_text) * 2)
                rom.write_ptr(base_text_address + nav_room.value * 8, text_addr)
                rom.write_ptr(base_text_address + nav_room.value * 8 + 4, text_addr)
                rom.write_16_list(text_addr, encoded_text)

    @classmethod
    def apply_hint_security(
        cls, rom: Rom, locks: dict[MarsschemaNavstationlocksKey, Hintlocks]
    ) -> None:
        """
        Applies an optional security level requirement to use Navigation Stations
        Defaults to OPEN if not provided in patch data JSON
        """
        default_lock_name = "OPEN"
        for location, offset in NavigationText.NAV_ROOM_ENUMS.items():
            rom.write_8(
                rom.read_ptr(ReservedConstants.HINT_SECURITY_LEVELS_ADDR) + offset.value,
                NavStationLockType[locks.get(location, default_lock_name)].value,
            )



import json
import typing
from collections.abc import Callable
from os import PathLike

from jsonschema import validate

from mars_patcher.auto_generated_types import MarsSchema
from mars_patcher.connections import Connections
from mars_patcher.credits import write_credits
from mars_patcher.data import get_data_path
from mars_patcher.door_locks import set_door_locks
from mars_patcher.item_patcher import ItemPatcher, set_required_metroid_count, set_tank_increments
from mars_patcher.level_edits import apply_level_edits
from mars_patcher.locations import LocationSettings
from mars_patcher.minimap import apply_base_minimap_edits, apply_minimap_edits
from mars_patcher.misc_patches import (
    apply_accessibility_patch,
    apply_base_patch,
    apply_pbs_without_bombs,
    apply_reveal_hidden_tiles,
    apply_reveal_unexplored_doors,
    apply_unexplored_map,
    change_missile_limit,
    disable_demos,
    disable_music,
    disable_sound_effects,
    skip_door_transitions,
    stereo_default,
)
from mars_patcher.navigation_text import NavigationText
from mars_patcher.random_palettes import PaletteRandomizer, PaletteSettings
from mars_patcher.rom import Rom
from mars_patcher.room_names import write_room_names
from mars_patcher.starting import set_starting_items, set_starting_location
from mars_patcher.text import write_seed_hash
from mars_patcher.titlescreen_text import write_title_text


def validate_patch_data(patch_data: dict) -> MarsSchema:
    """
    Validates whether the specified patch_data satisfies the schema for it.

    Raises:
        ValidationError: If the patch data does not satisfy the schema.
    """
    with open(get_data_path("schema.json")) as f:
        schema = json.load(f)
    validate(patch_data, schema)
    return typing.cast("MarsSchema", patch_data)


def patch(
    input_path: str | PathLike[str],
    output_path: str | PathLike[str],
    patch_data: MarsSchema,
    status_update: Callable[[str, float], None],
) -> None:
    """
    Creates a new randomized Fusion game, based off of an input path, an output path,
    a dictionary defining how the game should be randomized, and a status update function.

    Args:
        input_path: The path to an unmodified Metroid Fusion (U) ROM.
        output_path: The path where the randomized Fusion ROM should be saved to.
        patch_data: A dictionary defining how the game should be randomized.
            This function assumes that it satisfies the needed schema. To validate it, use
            validate_patch_data().
        status_update: A function taking in a message (str) and a progress value (float).
    """

    # Load input rom
    rom = Rom(input_path)

    # Apply base asm patch first
    apply_base_patch(rom)

    # Randomize palettes - palettes are randomized first in case the item
    # patcher needs to copy tilesets
    if "Palettes" in patch_data:
        status_update("Randomizing palettes...", -1)
        pal_settings = PaletteSettings.from_json(patch_data["Palettes"])
        pal_randomizer = PaletteRandomizer(rom, pal_settings)
        pal_randomizer.randomize()

    # Load locations and set assignments
    status_update("Writing item assignments...", -1)
    loc_settings = LocationSettings.initialize()
    loc_settings.set_assignments(patch_data["Locations"])
    item_patcher = ItemPatcher(rom, loc_settings)
    item_patcher.write_items()

    # Required metroid count
    set_required_metroid_count(rom, patch_data["RequiredMetroidCount"])

    # Starting location
    if "StartingLocation" in patch_data:
        status_update("Writing starting location...", -1)
        set_starting_location(rom, patch_data["StartingLocation"])

    # Starting items
    if "StartingItems" in patch_data:
        status_update("Writing starting items...", -1)
        set_starting_items(rom, patch_data["StartingItems"])

    # Tank increments
    if "TankIncrements" in patch_data:
        status_update("Writing tank increments...", -1)
        set_tank_increments(rom, patch_data["TankIncrements"])

    # Elevator connections
    conns = None
    if "ElevatorConnections" in patch_data:
        status_update("Writing elevator connections...", -1)
        conns = Connections(rom)
        conns.set_elevator_connections(patch_data["ElevatorConnections"])

    # Sector shortcuts
    if "SectorShortcuts" in patch_data:
        status_update("Writing sector shortcuts...", -1)
        if conns is None:
            conns = Connections(rom)
        conns.set_shortcut_connections(patch_data["SectorShortcuts"])

    # Hints
    if nav_text := patch_data.get("NavigationText", {}):
        status_update("Writing navigation text...", -1)
        navigation_text = NavigationText.from_json(nav_text)
        navigation_text.write(rom)

    if nav_locks := patch_data.get("NavStationLocks", {}):
        status_update("Writing navigation locks...", -1)
        NavigationText.apply_hint_security(rom, nav_locks)

    # Room Names
    if room_names := patch_data.get("RoomNames", []):
        status_update("Writing room names...", -1)
        write_room_names(rom, room_names)

    # Credits
    if credits_text := patch_data.get("CreditsText", []):
        status_update("Writing credits text...", -1)
        write_credits(rom, credits_text)

    # Misc patches
    if patch_data.get("AccessibilityPatches"):
        apply_accessibility_patch(rom)

    if patch_data.get("DisableDemos"):
        disable_demos(rom)

    if patch_data.get("SkipDoorTransitions"):
        skip_door_transitions(rom)

    if patch_data.get("StereoDefault", True):
        stereo_default(rom)

    if patch_data.get("DisableMusic"):
        disable_music(rom)

    if patch_data.get("DisableSoundEffects"):
        disable_sound_effects(rom)

    if "MissileLimit" in patch_data:
        change_missile_limit(rom, patch_data["MissileLimit"])

    if patch_data.get("PowerBombsWithoutBombs"):
        apply_pbs_without_bombs(rom)

    if patch_data.get("UnexploredMap"):
        apply_unexplored_map(rom)

        if not patch_data.get("HideDoorsOnMinimap", False):
            apply_reveal_unexplored_doors(rom)

    if patch_data.get("RevealHiddenTiles"):
        apply_reveal_hidden_tiles(rom)

    if "LevelEdits" in patch_data:
        apply_level_edits(rom, patch_data["LevelEdits"])

    # Apply base minimap edits
    apply_base_minimap_edits(rom)

    # Apply JSON minimap edits
    if "MinimapEdits" in patch_data:
        apply_minimap_edits(rom, patch_data["MinimapEdits"])

    # Door locks
    if door_locks := patch_data.get("DoorLocks", []):
        status_update("Writing door locks...", -1)
        set_door_locks(rom, door_locks)

    write_seed_hash(rom, patch_data["SeedHash"])

    # Title-screen text
    if title_screen_text := patch_data.get("TitleText"):
        status_update("Writing title screen text...", -1)
        write_title_text(rom, title_screen_text)

    rom.save(output_path)
    status_update(f"Output written to {output_path}", -1)



import random
from enum import Enum

import mars_patcher.constants.game_data as gd
from mars_patcher.auto_generated_types import (
    MarsschemaPalettes,
    MarsschemaPalettesColorspace,
    MarsschemaPalettesRandomize,
)
from mars_patcher.constants.palettes import (
    ENEMY_GROUPS,
    EXCLUDED_ENEMIES,
    MF_TILESET_ALT_PAL_ROWS,
    NETTORI_EXTRA_PALS,
    TILESET_ANIM_PALS,
)
from mars_patcher.palette import ColorChange, Palette, SineWave
from mars_patcher.rom import Game, Rom


class PaletteType(Enum):
    TILESETS = 1
    ENEMIES = 2
    SAMUS = 3
    BEAMS = 4


class PaletteSettings:
    PAL_TYPE_ENUMS = {
        "Tilesets": PaletteType.TILESETS,
        "Enemies": PaletteType.ENEMIES,
        "Samus": PaletteType.SAMUS,
        "Beams": PaletteType.BEAMS,
    }

    def __init__(
        self,
        seed: int,
        pal_types: dict[PaletteType, tuple[int, int]],  # TODO: change this tuple(int, int)
        color_space: MarsschemaPalettesColorspace,
        symmetric: bool,
        extra_variation: bool,
    ):
        self.seed = seed
        self.pal_types = pal_types
        self.color_space: MarsschemaPalettesColorspace = color_space
        self.symmetric = symmetric
        self.extra_variation = extra_variation

    @classmethod
    def from_json(cls, data: MarsschemaPalettes) -> "PaletteSettings":
        seed = data.get("Seed", random.randint(0, 2**31 - 1))
        random.seed(seed)
        pal_types = {}
        for type_name, hue_data in data["Randomize"].items():
            pal_type = cls.PAL_TYPE_ENUMS[type_name]
            hue_range = cls.get_hue_range(hue_data)
            pal_types[pal_type] = hue_range
        color_space = data.get("ColorSpace", "Oklab")
        symmetric = data.get("Symmetric", True)
        # Extra variation is always enabled. This could be passed via JSON instead.
        return cls(seed, pal_types, color_space, symmetric, True)

    @classmethod
    def get_hue_range(cls, data: MarsschemaPalettesRandomize) -> tuple[int, int]:
        hue_min = data.get("HueMin")
        hue_max = data.get("HueMax")
        if hue_min is None or hue_max is None:
            if hue_max is not None:
                hue_min = random.randint(0, hue_max)
            elif hue_min is not None:
                hue_max = random.randint(hue_min, 360)
            else:
                hue_min = random.randint(0, 360)
                hue_max = random.randint(hue_min, 360)
        if hue_min > hue_max:
            raise ValueError("HueMin cannot be greater than HueMax")
        return hue_min, hue_max


class PaletteRandomizer:
    """Class for randomly shifting the hues of color palettes."""

    def __init__(self, rom: Rom, settings: PaletteSettings):
        self.rom = rom
        self.settings = settings
        if settings.color_space == "HSV":
            self.change_func = self.change_palette_hsv
        elif settings.color_space == "Oklab":
            self.change_func = self.change_palette_oklab
        else:
            raise ValueError(f"Invalid color space '{settings.color_space}' for color space!")

    @staticmethod
    def change_palette_hsv(
        pal: Palette, change: ColorChange, excluded_rows: set[int] = set()
    ) -> None:
        pal.change_colors_hsv(change, excluded_rows)

    @staticmethod
    def change_palette_oklab(
        pal: Palette, change: ColorChange, excluded_rows: set[int] = set()
    ) -> None:
        pal.change_colors_oklab(change, excluded_rows)

    def generate_palette_change(self, hue_range: tuple[int, int]) -> ColorChange:
        """Generates a random color change. hue_range determines how far each color's hue will be
        initially rotated. Individual colors can be additionally rotated using the values of a
        random sine wave."""
        hue_shift = random.randint(hue_range[0], hue_range[1])
        if self.settings.symmetric and random.choice([True, False]):
            hue_shift = 360 - hue_shift
        if self.settings.extra_variation:
            hue_var_range = min(1.0, (hue_range[1] - hue_range[0]) / 180)
            hue_var = SineWave.generate(hue_var_range)
        else:
            hue_var = None
        return ColorChange(hue_shift, hue_var)

    def randomize(self) -> None:
        random.seed(self.settings.seed)
        self.randomized_pals: set[int] = set()
        pal_types = self.settings.pal_types
        if PaletteType.TILESETS in pal_types:
            self.randomize_tilesets(pal_types[PaletteType.TILESETS])
        if PaletteType.ENEMIES in pal_types:
            self.randomize_enemies(pal_types[PaletteType.ENEMIES])
        if PaletteType.SAMUS in pal_types:
            self.randomize_samus(pal_types[PaletteType.SAMUS])
        if PaletteType.BEAMS in pal_types:
            self.randomize_beams(pal_types[PaletteType.BEAMS])
        # Fix any sprite/tileset palettes that should be the same
        # TODO: Check for palette fixes needed in fusion
        if self.rom.is_zm():
            self.fix_zm_palettes()

    def change_palettes(self, pals: list[tuple[int, int]], change: ColorChange) -> None:
        for addr, rows in pals:
            if addr in self.randomized_pals:
                continue
            pal = Palette(rows, self.rom, addr)
            self.change_func(pal, change)
            pal.write(self.rom, addr)
            self.randomized_pals.add(addr)

    def randomize_samus(self, hue_range: tuple[int, int]) -> None:
        change = self.generate_palette_change(hue_range)
        self.change_palettes(gd.samus_palettes(self.rom), change)
        self.change_palettes(gd.helmet_cursor_palettes(self.rom), change)
        self.change_palettes(gd.sax_palettes(self.rom), change)

    def randomize_beams(self, hue_range: tuple[int, int]) -> None:
        change = self.generate_palette_change(hue_range)
        self.change_palettes(gd.beam_palettes(self.rom), change)

    def randomize_tilesets(self, hue_range: tuple[int, int]) -> None:
        rom = self.rom
        ts_addr = gd.tileset_entries(rom)
        ts_count = gd.tileset_count(rom)
        anim_pal_count = gd.anim_palette_count(rom)
        anim_pal_to_randomize = set(range(anim_pal_count))

        for _ in range(ts_count):
            # Get tileset palette address
            pal_ptr = ts_addr + 4
            pal_addr = rom.read_ptr(pal_ptr)
            ts_addr += 0x14
            if pal_addr in self.randomized_pals:
                continue
            # Get excluded palette rows
            excluded_rows = set()
            if rom.game == Game.MF:
                row = MF_TILESET_ALT_PAL_ROWS.get(pal_addr)
                if row is not None:
                    excluded_rows = {row}
            # Load palette and shift hue
            pal = Palette(13, rom, pal_addr)
            change = self.generate_palette_change(hue_range)
            self.change_func(pal, change, excluded_rows)
            pal.write(rom, pal_addr)
            self.randomized_pals.add(pal_addr)
            # Check animated palette
            anim_pal_id = TILESET_ANIM_PALS.get(pal_addr)
            if anim_pal_id is not None:
                self.randomize_anim_palette(anim_pal_id, change)
                anim_pal_to_randomize.remove(anim_pal_id)

        # Go through remaining animated palettes
        for anim_pal_id in anim_pal_to_randomize:
            change = self.generate_palette_change(hue_range)
            self.randomize_anim_palette(anim_pal_id, change)

    def randomize_anim_palette(self, anim_pal_id: int, change: ColorChange) -> None:
        rom = self.rom
        addr = gd.anim_palette_entries(rom) + anim_pal_id * 8
        pal_addr = rom.read_ptr(addr + 4)
        if pal_addr in self.randomized_pals:
            return
        rows = rom.read_8(addr + 2)
        pal = Palette(rows, rom, pal_addr)
        self.change_func(pal, change)
        pal.write(rom, pal_addr)
        self.randomized_pals.add(pal_addr)

    def randomize_enemies(self, hue_range: tuple[int, int]) -> None:
        rom = self.rom
        excluded = EXCLUDED_ENEMIES[rom.game]
        sp_count = gd.sprite_count(rom)
        to_randomize = set(range(0x10, sp_count))
        to_randomize -= excluded

        # Go through sprites in groups
        groups = ENEMY_GROUPS[rom.game]
        for _, sprite_ids in groups.items():
            change = self.generate_palette_change(hue_range)
            for sprite_id in sprite_ids:
                assert sprite_id in to_randomize, f"{sprite_id:X} should be excluded"
                self.randomize_enemy(sprite_id, change)
                to_randomize.remove(sprite_id)

        # Go through remaining sprites
        for sprite_id in to_randomize:
            change = self.generate_palette_change(hue_range)
            self.randomize_enemy(sprite_id, change)

    def randomize_enemy(self, sprite_id: int, change: ColorChange) -> None:
        # Get palette address and row count
        rom = self.rom
        sprite_gfx_id = sprite_id - 0x10
        pal_ptr = gd.sprite_palette_ptrs(rom)
        pal_addr = rom.read_ptr(pal_ptr + sprite_gfx_id * 4)
        if pal_addr in self.randomized_pals:
            return
        if rom.is_mf():
            if sprite_id == 0x4D or sprite_id == 0xBE:
                # Ice beam ability and zozoros only have 1 row, not 2
                rows = 1
            else:
                vram_size_addr = gd.sprite_vram_sizes(rom)
                vram_size = rom.read_32(vram_size_addr + sprite_gfx_id * 4)
                rows = vram_size // 0x800
        elif rom.is_zm():
            gfx_ptr = gd.sprite_graphics_ptrs(rom)
            gfx_addr = rom.read_ptr(gfx_ptr + sprite_gfx_id * 4)
            rows = (rom.read_32(gfx_addr) >> 8) // 0x800
        else:
            raise ValueError("Unknown game!")
        # Load palette, change colors, and write to ROM
        pal = Palette(rows, rom, pal_addr)
        self.change_func(pal, change)
        pal.write(rom, pal_addr)
        self.randomized_pals.add(pal_addr)
        if rom.is_mf() and sprite_id == 0x26:
            self.fix_nettori(change)

    def get_sprite_addr(self, sprite_id: int) -> int:
        addr = gd.sprite_palette_ptrs(self.rom) + (sprite_id - 0x10) * 4
        return self.rom.read_ptr(addr)

    def get_tileset_addr(self, sprite_id: int) -> int:
        addr = gd.tileset_entries(self.rom) + sprite_id * 0x14 + 4
        return self.rom.read_ptr(addr)

    def fix_nettori(self, change: ColorChange) -> None:
        """Nettori has extra palettes stored separately, so they require the same color change."""
        for addr, rows in NETTORI_EXTRA_PALS:
            pal = Palette(rows, self.rom, addr)
            self.change_func(pal, change)
            pal.write(self.rom, addr)

    def fix_zm_palettes(self) -> None:
        if (
            PaletteType.ENEMIES in self.settings.pal_types
            or PaletteType.TILESETS in self.settings.pal_types
        ):
            # Fix kraid's body
            sp_addr = self.get_sprite_addr(0x6F)
            ts_addr = self.get_tileset_addr(9)
            self.rom.copy_bytes(sp_addr, ts_addr + 0x100, 0x20)

        if PaletteType.TILESETS in self.settings.pal_types:
            # Fix kraid elevator statue
            sp_addr = self.get_sprite_addr(0x95)
            ts_addr = self.get_tileset_addr(0x35)
            self.rom.copy_bytes(ts_addr + 0x20, sp_addr, 0x20)

            # Fix ridley elevator statue
            ts_addr = self.get_tileset_addr(7)
            self.rom.copy_bytes(ts_addr + 0x20, sp_addr + 0x20, 0x20)

            # Fix tourian statues
            sp_addr = self.get_sprite_addr(0xA3)
            ts_addr = self.get_tileset_addr(0x41)
            self.rom.copy_bytes(ts_addr + 0x60, sp_addr, 0x20)
            # Fix cutscene
            sp_addr = gd.tourian_statues_cutscene_palette(self.rom)
            self.rom.copy_bytes(ts_addr, sp_addr, 0xC0)



from mars_patcher.rom import Rom
from mars_patcher.room_entry import RoomEntry


def apply_level_edits(rom: Rom, edit_dict: dict) -> None:
    # Go through every area
    for area, rooms in edit_dict.items():
        # Go through every room
        for room, layers in rooms.items():
            r = RoomEntry(rom, int(area), int(room))

            # Go through every layer
            for layer, changes in layers.items():
                if layer == "BG1":
                    load = r.load_bg1
                elif layer == "BG2":
                    load = r.load_bg2
                elif layer == "Clipdata":
                    load = r.load_clip
                else:
                    raise ValueError("Unsupported Block Layer")

                # Load layer, do every edit that's provided and write back.
                with load() as layer:
                    for change in changes:
                        layer.set_block_value(change["X"], change["Y"], change["Value"])



from mars_patcher.auto_generated_types import MarsschemaStartingitems, MarsschemaStartinglocation
from mars_patcher.constants.game_data import area_doors_ptrs, spriteset_ptrs, starting_equipment
from mars_patcher.constants.items import BEAM_FLAGS, MISSILE_BOMB_FLAGS, SUIT_MISC_FLAGS
from mars_patcher.constants.reserved_space import ReservedConstants
from mars_patcher.rom import Rom
from mars_patcher.room_entry import RoomEntry

# Keep in sync with base patch
STARTING_LOC_ADDR = ReservedConstants.STARTING_LOCATION_ADDR


def set_starting_location(rom: Rom, data: MarsschemaStartinglocation) -> None:
    area = data["Area"]
    room = data["Room"]
    # Don't do anything for area 0 room 0
    if area == 0 and room == 0:
        return
    # Find any door in the provided room
    door = find_door_in_room(rom, area, room)
    # Check if save pad in room
    pos = find_save_pad_position(rom, area, room)
    if pos is not None:
        x_pos, y_pos = pos
    else:
        # Convert block coordinates to actual position
        x_pos = data["BlockX"] * 64 + 31
        y_pos = data["BlockY"] * 64 + 63
    # Write to rom
    starting_location = rom.read_ptr(STARTING_LOC_ADDR)
    rom.write_8(starting_location, area)
    rom.write_8(starting_location + 1, room)
    rom.write_8(starting_location + 2, door)
    rom.write_16(starting_location + 4, x_pos)
    rom.write_16(starting_location + 6, y_pos)


def find_door_in_room(rom: Rom, area: int, room: int) -> int:
    door_addr = rom.read_ptr(area_doors_ptrs(rom) + area * 4)
    door = None
    for d in range(256):
        if rom.read_8(door_addr) == 0:
            break
        if rom.read_8(door_addr + 1) == room:
            door = d
            break
        door_addr += 0xC
    if door is None:
        raise ValueError(f"No door found for area {area} room {room:X}")
    return door


def find_save_pad_position(rom: Rom, area: int, room: int) -> tuple[int, int] | None:
    # Check if room's spriteset has save pad
    room_entry = RoomEntry(rom, area, room)
    spriteset = room_entry.default_spriteset()
    ss_addr = rom.read_ptr(spriteset_ptrs(rom) + spriteset * 4)
    ss_idx = None
    for i in range(15):
        sprite_id = rom.read_8(ss_addr)
        if sprite_id == 0:
            break
        if sprite_id == 0x1F:
            ss_idx = i
            break
        ss_addr += 2
    if ss_idx is None:
        return None
    # Find save pad in sprite layout list
    layout_addr = room_entry.default_sprite_layout_addr()
    for i in range(24):
        sp_y = rom.read_8(layout_addr)
        sp_x = rom.read_8(layout_addr + 1)
        prop = rom.read_8(layout_addr + 2)
        if sp_x == 0xFF and sp_y == 0xFF and prop == 0xFF:
            break
        if (prop & 0xF) - 1 == ss_idx:
            x_pos = sp_x * 64 + 32
            y_pos = sp_y * 64 + 9
            return x_pos, y_pos
        layout_addr += 3
    # No save pad found
    return None


def set_starting_items(rom: Rom, data: MarsschemaStartingitems) -> None:
    def get_ability_flags(ability_flags: dict[str, int]) -> int:
        status = 0
        for ability, flag in ability_flags.items():
            if ability in abilities:
                status |= flag
        return status

    # Get health/ammo amounts
    energy = data.get("Energy", 99)
    missiles = data.get("Missiles", 10)
    power_bombs = data.get("PowerBombs", 10)
    # Get ability status flags
    abilities = data.get("Abilities", [])
    beam_status = get_ability_flags(BEAM_FLAGS)
    missile_bomb_status = get_ability_flags(MISSILE_BOMB_FLAGS)
    suit_misc_status = get_ability_flags(SUIT_MISC_FLAGS)
    # Get security level flags
    levels = data.get("SecurityLevels", [0])
    level_status = 0
    for level in levels:
        level_status |= 1 << level
    # Get downloaded map flags
    maps = data.get("DownloadedMaps", range(7))
    map_status = 0
    for map in maps:
        map_status |= 1 << map
    # Write to rom
    addr = starting_equipment(rom)
    rom.write_16(addr, energy)
    rom.write_16(addr + 2, energy)
    rom.write_16(addr + 4, missiles)
    rom.write_16(addr + 6, missiles)
    rom.write_8(addr + 8, power_bombs)
    rom.write_8(addr + 9, power_bombs)
    rom.write_8(addr + 0xA, beam_status)
    rom.write_8(addr + 0xB, missile_bomb_status)
    rom.write_8(addr + 0xC, suit_misc_status)
    rom.write_8(addr + 0xD, level_status)
    rom.write_8(addr + 0xE, map_status)
